import unittest

from .chord_parser import ChordParser

from ..keys.chords import (
    PowerChord, Suspended4Chord, Suspended2Chord,
    MajorTriad, MinorTriad, AugmentedTriad, DiminishedTriad,
    MajorSixthChord, MinorSixthChord,
    MajorSeventhChord, MinorSeventhChord, DominantSeventhChord,
    HalfDiminishedSeventhChord, DiminishedSeventhChord,
    MajorAdd9Chord, MinorAdd9Chord, AugmentedAdd9Chord, DiminishedAdd9Chord,
    DominantNinthChord, DominantMinorNinthChord,
    MajorNinthChord, MinorNinthChord,
)

from ..notes.constants import (
    C_3, F_3, E_3, A_3, D_3,
    B_FLAT_3, E_FLAT_3, D_FLAT_3, C_FLAT_3, B_FLAT_3,
    A_SHARP_3, C_SHARP_3, E_SHARP_3,
    G_DOUBLE_FLAT_3,
    G_DOUBLE_SHARP_3,
)

from ..errors import InvalidChord

__all__ = [
    'ChordParserTest',
    'NonThirdChordsTest',
    'TriadChordsTest',
    'SixthChordsTest',
    'SeventhChordsTest',
    'Add9ChordsTest',
    'NinthChordsTest',
]


class ChordParserTest(unittest.TestCase):

    def testInvalidChords(self):
        self.assertRaises(InvalidChord, ChordParser('').parse)
        self.assertRaises(InvalidChord, ChordParser('#').parse)
        self.assertRaises(InvalidChord, ChordParser('♯').parse)
        self.assertRaises(InvalidChord, ChordParser('♭').parse)
        self.assertRaises(InvalidChord, ChordParser('#F').parse)
        self.assertRaises(InvalidChord, ChordParser('♭B').parse)
        self.assertRaises(InvalidChord, ChordParser('H').parse)
        self.assertRaises(InvalidChord, ChordParser('h').parse)

        self.assertRaises(InvalidChord, ChordParser('Dsharp').parse)
        self.assertRaises(InvalidChord, ChordParser('Eflat').parse)
        self.assertRaises(InvalidChord, ChordParser('pmin7').parse)

        # self.assertRaises(InvalidChord, ChordParser('CB').parse)  # should fail...
        # self.assertRaises(InvalidChord, ChordParser('Cqwe').parse)  # should fail...


class NonThirdChordsTest(unittest.TestCase):

    def testPowerChordClasses(self):
        assert isinstance(ChordParser('g5').parse(), PowerChord)

    def testPowerChordRoots(self):
        assert ChordParser('F5').parse().root ** F_3


    def testSuspended4ChordClasses(self):
        assert isinstance(ChordParser('D♭♭sus4').parse(), Suspended4Chord)
        assert isinstance(ChordParser('Fsus').parse(), Suspended4Chord)

    def testSuspended4ChordRoots(self):
        assert ChordParser('G𝄫sus4').parse().root ** G_DOUBLE_FLAT_3
        assert ChordParser('gbbsus').parse().root ** G_DOUBLE_FLAT_3


    def testSuspended2ChordClasses(self):
        assert isinstance(ChordParser('A♯sus2').parse(), Suspended2Chord)
        assert isinstance(ChordParser('Esus9').parse(), Suspended2Chord)

    def testSuspended4ChordRoots(self):
        assert ChordParser('G##sus2').parse().root ** G_DOUBLE_SHARP_3
        assert ChordParser('g𝄪sus9').parse().root ** G_DOUBLE_SHARP_3




class TriadChordsTest(unittest.TestCase):

    def testMajorTriadClasses(self):
        assert isinstance(ChordParser('A').parse(), MajorTriad)
        assert isinstance(ChordParser('Dmaj').parse(), MajorTriad)
        assert isinstance(ChordParser('Emajor').parse(), MajorTriad)

    def testMajorTriadRoots(self):
        assert ChordParser('C').parse().root ** C_3
        assert ChordParser('Cmaj').parse().root ** C_3
        assert ChordParser('Cmajor').parse().root ** C_3

    # def testMajorTriadBass(self):
    #     assert ChordParser('C').parse().bass ** C_3
    #     assert ChordParser('Cmaj').parse().bass ** C_3
    #     assert ChordParser('Cmajor').parse().bass ** C_3


    def testMinorTriadClasses(self):
        assert isinstance(ChordParser('Bmin').parse(), MinorTriad)
        assert isinstance(ChordParser('F-').parse(), MinorTriad)
        assert isinstance(ChordParser('Cminor').parse(), MinorTriad)

    def testMinorTriadRoots(self):
        assert ChordParser('Emin').parse().root ** E_3
        assert ChordParser('e-').parse().root ** E_3
        assert ChordParser('Eminor').parse().root ** E_3


    def testAugTriadClasses(self):
        assert isinstance(ChordParser('C##aug').parse(), AugmentedTriad)
        assert isinstance(ChordParser('Bbaugmented').parse(), AugmentedTriad)

    def testAugTriadRoots(self):
        assert ChordParser('Aaug').parse().root ** A_3
        assert ChordParser('Aaugmented').parse().root ** A_3


    def testDimTriadClasses(self):
        assert isinstance(ChordParser('D#dim').parse(), DiminishedTriad)
        assert isinstance(ChordParser('Bbdiminished').parse(), DiminishedTriad)

    def testDimTriadRoots(self):
        assert ChordParser('Ddim').parse().root ** D_3
        assert ChordParser('Ddiminished').parse().root ** D_3
        assert ChordParser('D dim').parse().root ** D_3


class SixthChordsTest(unittest.TestCase):

    def testMajorSixthChordClasses(self):
        assert isinstance(ChordParser('D#6').parse(), MajorSixthChord)
        assert isinstance(ChordParser('Fadd6').parse(), MajorSixthChord)

    def testMajorSixthChordRoots(self):
        assert ChordParser('Bb6').parse().root ** B_FLAT_3
        assert ChordParser('B♭add6').parse().root ** B_FLAT_3


    def testMinorSixthChordClasses(self):
        assert isinstance(ChordParser('Ebm6').parse(), MinorSixthChord)
        assert isinstance(ChordParser('Fmin6').parse(), MinorSixthChord)

    def testMinorSixthChordRoots(self):
        assert ChordParser('A♯m6').parse().root ** A_SHARP_3
        assert ChordParser('A#min6').parse().root ** A_SHARP_3


class SeventhChordsTest(unittest.TestCase):

    def testMajorSeventhChordClasses(self):
        assert isinstance(ChordParser('Amaj7').parse(), MajorSeventhChord)
        assert isinstance(ChordParser('BM7').parse(), MajorSeventhChord)
        assert isinstance(ChordParser('CΔ7').parse(), MajorSeventhChord)
        assert isinstance(ChordParser('Dmajor7').parse(), MajorSeventhChord)

    def testMajorSeventhChordRoots(self):
        assert ChordParser('Cmaj7').parse().root ** C_3
        assert ChordParser('CM7').parse().root ** C_3
        assert ChordParser('CΔ7').parse().root ** C_3
        assert ChordParser('Cmajor7').parse().root ** C_3


    def testMinorSeventhChordClasses(self):
        assert isinstance(ChordParser('Emin7').parse(), MinorSeventhChord)
        assert isinstance(ChordParser('Fm7').parse(), MinorSeventhChord)
        assert isinstance(ChordParser('G-7').parse(), MinorSeventhChord)
        assert isinstance(ChordParser('Bbminor7').parse(), MinorSeventhChord)

    def testMinorSeventhChordRoots(self):
        assert ChordParser('Amin7').parse().root ** A_3
        assert ChordParser('Am7').parse().root ** A_3
        assert ChordParser('A-7').parse().root ** A_3
        assert ChordParser('Aminor7').parse().root ** A_3
        assert ChordParser('A minor7').parse().root ** A_3


    def testDomSeventhChordClasses(self):
        assert isinstance(ChordParser('F7').parse(), DominantSeventhChord)
        assert isinstance(ChordParser('Bbdom7').parse(), DominantSeventhChord)
        assert isinstance(ChordParser('Dbbdominant7').parse(), DominantSeventhChord)

    def testDomSeventhChordRoots(self):
        assert ChordParser('C#7').parse().root ** C_SHARP_3
        assert ChordParser('C♯dom7').parse().root ** C_SHARP_3
        assert ChordParser('c#dominant7').parse().root ** C_SHARP_3


    def testHalfDiminishedSeventhChordClasses(self):
        assert isinstance(ChordParser('Em7b5').parse(), HalfDiminishedSeventhChord)
        assert isinstance(ChordParser('F#m7-5').parse(), HalfDiminishedSeventhChord)
        assert isinstance(ChordParser('Gmin7dim5').parse(), HalfDiminishedSeventhChord)
        assert isinstance(ChordParser('Gbm7(b5)').parse(), HalfDiminishedSeventhChord)
        assert isinstance(ChordParser('fø7').parse(), HalfDiminishedSeventhChord)

    def testHalfDiminishedSeventhChordRoots(self):
        assert ChordParser('ebm7b5').parse().root ** E_FLAT_3
        assert ChordParser('E♭m7-5').parse().root ** E_FLAT_3
        assert ChordParser('Ebmin7dim5').parse().root ** E_FLAT_3
        assert ChordParser('Ebm7(b5)').parse().root ** E_FLAT_3
        assert ChordParser('Ebø7').parse().root ** E_FLAT_3


    def testDiminishedSeventhChordClasses(self):
        assert isinstance(ChordParser('fdim7').parse(), DiminishedSeventhChord)
        assert isinstance(ChordParser('ao7').parse(), DiminishedSeventhChord)
        assert isinstance(ChordParser('Cdiminished7').parse(), DiminishedSeventhChord)

    def testDiminishedSeventhChordRoots(self):
        assert ChordParser('Dbdim7').parse().root ** D_FLAT_3
        assert ChordParser('Dbo7').parse().root ** D_FLAT_3
        assert ChordParser('Dbdiminished7').parse().root ** D_FLAT_3


class Add9ChordsTest(unittest.TestCase):

    def testMajorAdd9ChordClasses(self):
        assert isinstance(ChordParser('bAdd9').parse(), MajorAdd9Chord)
        assert isinstance(ChordParser('Aadd9').parse(), MajorAdd9Chord)

    def testMajorAdd9ChordRoots(self):
        assert ChordParser('Dbadd9').parse().root ** D_FLAT_3
        assert ChordParser('d♭Add9').parse().root ** D_FLAT_3


    def testMinorAdd9ChordClasses(self):
        assert isinstance(ChordParser('Emadd9').parse(), MinorAdd9Chord)
        assert isinstance(ChordParser('GmAdd9').parse(), MinorAdd9Chord)

    def testMinorAdd9ChordRoots(self):
        assert ChordParser('Cbmadd9').parse().root ** C_FLAT_3
        assert ChordParser('C♭mAdd9').parse().root ** C_FLAT_3


    def testAugmentedAdd9ChordClasses(self):
        assert isinstance(ChordParser('f#augadd9').parse(), AugmentedAdd9Chord)
        assert isinstance(ChordParser('D#augAdd9').parse(), AugmentedAdd9Chord)

    def testAugmentedAdd9ChordRoots(self):
        assert ChordParser('E♯augadd9').parse().root ** E_SHARP_3
        assert ChordParser('E#augAdd9').parse().root ** E_SHARP_3


    def testDiminishedAdd9ChordClasses(self):
        assert isinstance(ChordParser('Cdimadd9').parse(), DiminishedAdd9Chord)
        assert isinstance(ChordParser('F#dimAdd9').parse(), DiminishedAdd9Chord)

    def testDiminishedAdd9ChordRoots(self):
        assert ChordParser('E♯dimadd9').parse().root ** E_SHARP_3
        assert ChordParser('E#dimAdd9').parse().root ** E_SHARP_3


class NinthChordsTest(unittest.TestCase):

    def testMajorNinthChordClasses(self):
        assert isinstance(ChordParser('D#maj9').parse(), MajorNinthChord)
        assert isinstance(ChordParser('FM9').parse(), MajorNinthChord)
        assert isinstance(ChordParser('Emajor9').parse(), MajorNinthChord)

    def testMajorNinthChordRoots(self):
        assert ChordParser('Bbmaj7').parse().root ** B_FLAT_3
        assert ChordParser('B♭M7').parse().root ** B_FLAT_3
        assert ChordParser('b♭major7').parse().root ** B_FLAT_3


    def testMinorNinthChordClasses(self):
        assert isinstance(ChordParser('Ebmin9').parse(), MinorNinthChord)
        assert isinstance(ChordParser('Fm9').parse(), MinorNinthChord)
        assert isinstance(ChordParser('G#-9').parse(), MinorNinthChord)
        assert isinstance(ChordParser('Bminor9').parse(), MinorNinthChord)

    def testMinorNinthChordRoots(self):
        assert ChordParser('A♯min9').parse().root ** A_SHARP_3
        assert ChordParser('A#m9').parse().root ** A_SHARP_3
        assert ChordParser('A#-9').parse().root ** A_SHARP_3
        assert ChordParser('A♯minor9').parse().root ** A_SHARP_3
        assert ChordParser('A# min9').parse().root ** A_SHARP_3


    def testDomNinthChordClasses(self):
        assert isinstance(ChordParser('F9').parse(), DominantNinthChord)
        assert isinstance(ChordParser('Bbdom9').parse(), DominantNinthChord)

    def testDomNinthChordRoots(self):
        assert ChordParser('C#9').parse().root ** C_SHARP_3
        assert ChordParser('C♯dom9').parse().root ** C_SHARP_3


    def testDomMinNinthChordClasses(self):
        assert isinstance(ChordParser('E7b9').parse(), DominantMinorNinthChord)

    def testDomMinNinthChordRoots(self):
        assert ChordParser('eb7b9').parse().root ** E_FLAT_3






__all__ = [
    'InvalidInstrument',
    'InvalidScale',
    'InvalidChord',
    'InvalidNote',
    'InvalidAlteration',
    'InvalidOctave',
]


class InvalidInstrument(Exception):
    pass

class InvalidScale(Exception):
    pass

class InvalidChord(Exception):
    pass

class InvalidNote(Exception):
    pass

class InvalidAlteration(Exception):
    pass

class InvalidOctave(Exception):
    pass



import unittest

from .pitch_parser import NotePitchParser

from ..notes import NotePitch

from ..notes.constants import (
    C_3, F_3, E_3, A_3, D_3, G_3, B_3,
    D_SHARP_3, E_FLAT_3, C_DOUBLE_SHARP_3, G_DOUBLE_FLAT_3,
)

from ..errors import InvalidNote, InvalidAlteration, InvalidOctave

__all__ = [
    'NotePitchParserTest',
]

class NotePitchParserTest(unittest.TestCase):

    CHAR_WINS = [
        ['C', C_3],
        ['D', D_3],
        ['E', E_3],
        ['F', F_3],
        ['G', G_3],
        ['A', A_3],
        ['B', B_3],

        ['c', C_3],
        ['d', D_3],
        ['e', E_3],
        ['f', F_3],
        ['g', G_3],
        ['a', A_3],
        ['b', B_3],
    ]

    CHAR_FAILS = [
        '',
        'H',
        'T',
        'Y',
        'h',
    ]

    OCTAVE_WINS = [
        [ f'C{octave}', NotePitch('C', octave) ] for octave
        in range(0, NotePitch.MAXIMUM_OCTAVE)
    ]

    OCTAVE_FAILS = [
        f'A{octave}' for octave
        in range(NotePitch.MAXIMUM_OCTAVE+1, NotePitch.MAXIMUM_OCTAVE+500)
    ]

    ALTS_WINS = [
        ['D#', D_SHARP_3],
        ['D♯', D_SHARP_3],

        ['C##', C_DOUBLE_SHARP_3],
        ['C♯♯', C_DOUBLE_SHARP_3],
        ['C#♯', C_DOUBLE_SHARP_3],
        ['C♯#', C_DOUBLE_SHARP_3],
        ['C𝄪',  C_DOUBLE_SHARP_3],

        ['Eb', E_FLAT_3],
        ['E♭', E_FLAT_3],

        ['Gbb', G_DOUBLE_FLAT_3],
        ['G♭♭', G_DOUBLE_FLAT_3],
        ['Gb♭', G_DOUBLE_FLAT_3],
        ['G♭b', G_DOUBLE_FLAT_3],
        ['G𝄫',  G_DOUBLE_FLAT_3],
    ]

    ALTS_FAILS = [
        'Abbb',
        'A###',
        'Bbr6',
        'Ce#',

        'D𝄫b',
        'D𝄫♭',

        'G𝄪#',
        'G𝄪♯',

        'Dsharp',
        'eSharp',
        'Eflat',
        'bFlat',

        'CB',
        'Cqwe',
    ]

    def setUp(self):
        print()

    def testChars(self):
        for symbol, expected in self.CHAR_WINS:
            parser = NotePitchParser(symbol)
            parsed_note = parser.parse()
            self.assertEqual(parsed_note, expected)

    def testCharFails(self):
        for symbol in self.CHAR_FAILS:
            parser = NotePitchParser(symbol)
            self.assertRaises(InvalidNote, parser.parse)

    def testOctaves(self):
        for symbol, expected in self.OCTAVE_WINS:
            parser = NotePitchParser(symbol)
            parsed_note = parser.parse()
            self.assertEqual(parsed_note, expected)

    def testOctaveFails(self):
        for symbol in self.OCTAVE_FAILS:
            parser = NotePitchParser(symbol)
            self.assertRaises(InvalidOctave, parser.parse)

    def testAlterations(self):
        for symbol, expected in self.ALTS_WINS:
            parser = NotePitchParser(symbol)
            parsed_note = parser.parse()
            self.assertEqual(parsed_note, expected)

    def testAlterationFails(self):
        for symbol in self.ALTS_FAILS:
            parser = NotePitchParser(symbol)
            self.assertRaises(InvalidAlteration, parser.parse)



'''
㉑ ㉒ ㉓ ㉔ ㉕ ㉖ ㉗ ㉘ ㉙ ㉚
㉛ ㉜ ㉝ ㉞ ㉟ ㊱ ㊲ ㊳ ㊴ ㊵
㊶ ㊷ ㊸ ㊹ ㊺ ㊻ ㊼ ㊽ ㊾ ㊿

These are specifically sans-serif:

🄋 ➀ ➁ ➂ ➃ ➄ ➅ ➆ ➇ ➈ ➉

Black Circled Number
⓿
❶
❷
❸
❹
❺
❻
❼
❽
❾
❿


Ⓐ ⓐ
Ⓑ ⓑ
Ⓒ ⓒ
Ⓓ ⓓ
Ⓔ ⓔ
Ⓕ ⓕ
Ⓖ ⓖ


# NUMERALS = {
#     'I' : 'Ⅰ',
#     'V' : 'ⅤⅠⅤ',
#     'X' : 'Ⅹ',
#     'L' : 'Ⅼ',
#     'C' : 'Ⅽ',
#     'D' : 'Ⅾ',
#     'M' : 'Ⅿ',
# }

https://www.unicode.org/charts/nameslist/n_2460.html

'''

from bestia.output import Row, FString, echo, tty_cols

from .keys.scales import ChromaticScale
from .notes import NotePitch

__all__ = [
    'PluckedStringInstrument',
    'max_frets_on_screen',
]


class PluckedString:

    DEGREE_ICONS = (
        '⓪',  # null degree...
        # 'Ⓝ',
        # '①',
        'Ⓡ',
        '➁',
        '➂',
        '➃',
        '➄',
        '➅',
        '➆',
        '➇',
        '➈',
        '➉',
        '⑪',
        '⑫',
        '⑬',
        '⑭',
        '⑮',
        '⑯',
        '⑰',
        '⑱',
        '⑲',
        '⑳',
    )

    def __init__(
        self,
        char, alt='', oct=NotePitch.DEFAULT_OCTAVE,
        frets=12,
        mode=None,
        verbose=1,
        show_degrees=False
    ):
        self.tuning = NotePitch(char, alt, oct)
        self.mode = mode
        self.frets = frets
        self.verbose = verbose
        self.show_degrees = show_degrees

    def __repr__(self):
        ''' prints string notes matching given key '''
        string_line = Row()

        mode = self.mode if self.mode else ChromaticScale(*self.tuning)

        for f, note in enumerate(
            mode.spell(
                note_count=self.frets + 1,
                start_note=self.tuning,
                yield_all=True,
            )
        ):

            fret_value = ''
            if note:

                note_fg = 'green' if note ** mode.root else 'magenta'

                if self.show_degrees:
                    for d in mode.allowed_degrees():
                        if note ** mode[d]:
                            fret_value = FString(
                                '{} '.format(
                                    self.DEGREE_ICONS[d] if self.verbose != 0
                                    else (
                                        d if d != 1 else 'R'
                                    )
                                ),
                                size=3,
                                fg=note_fg,
                                align='r',
                            )
                            break

                else:
                    fret_value = '{}{}{}'.format(
                        FString(
                            note.chr,
                            size=1,
                            fg=note_fg,
                            fx=[''],
                        ),
                        FString(
                            note.repr_alt,
                            size=0,
                            fg=note_fg,
                            fx=[''],
                        ),
                        FString(
                            note.repr_oct if self.verbose > 0 else '',
                            size=1,
                            fg=note_fg,
                            fx=['faint'],
                        ),
                    )


            # APPEND NOTE PITCH DATA
            string_line.append(
                FString(
                    fret_value,
                    size=PluckedStringInstrument.NOTE_WIDTH,
                    align='cr',
                )
            )

            # APPEND FRET SYMBOL
            string_line.append(
                FString(
                    '║' if f % 12 == 0 else '│',
                    size=PluckedStringInstrument.FRET_WIDTH,
                    # fg='blue',
                    # fx=['faint'],
                )
            )

            if f == self.frets:
                break

        return str(string_line)


class PluckedStringInstrument:

    NOTE_WIDTH = 5
    FRET_WIDTH = 1

    _INLAYS = (
        'I',
        'II',
        'III',
        'IV',
        'V',
        'VI',
        'VII',
        'VIII',
        'IX',
        'X',
        'XI',

        'XII',
        'XIII',
        'XIV',
        'XV',
        'XVI',
        'XVII',
        'XVIII',
        'XIX',
        'XX',
        'XXI',
        'XXII',
        'XXIII',

        'XXIV',
        'XXV',
        'XXVI',
        'XXVII',
        'XXVIII',
        'XXIX',
        'XXX',
        'XXXI',
        'XXXII',
        'XXXIII',
        'XXXIV',
        'XXXV',
        'XXXVI',
    )

    @classmethod
    def maximum_frets(cls):
        return len(cls._INLAYS)


    @classmethod
    def total_fret_width(cls):
        return cls.NOTE_WIDTH + cls.FRET_WIDTH


    def __init__(self, *notes, name=''):
        self.strings = [ PluckedString(*n) for n in notes ]
        self.name = name


    @property
    def string_n_size(self):
        ''' does not expect more then 99 strings... '''
        return 1 if len(self.strings) < 10 else 2


    def render_inlays(self, frets=12, verbose=1):

        if not verbose:
            return

        dots = (3, 5, 7, 9, 12, 15, 17, 19, 21, 24, 27, 29, 31, 33, 36)

        inlay_row = Row(
            FString(
                '',
                size=self.string_n_size + self.total_fret_width(),
                align='l',
                # pad='*',
            ),
        )

        f = 1
        while frets:
            inlay_row.append(
                FString(
                    '' if verbose == 1 and f not in dots else self._INLAYS[f-1],
                    size=self.total_fret_width(),
                    align='r' if verbose == 2 else 'c', # high verbose needs more space
                    fg='cyan',
                    fx=['faint' if f not in dots else ''],
                )
            )
            f += 1
            frets -= 1

        echo(inlay_row)


    def render_binding(self, side, frets=12):
        ''' https://en.wikipedia.org/wiki/Box-drawing_character
        '''
        normal = { 'upper': '═', 'lower': '═' }
        twelve = { 'upper': '╦', 'lower': '╩' }
        joints = { 'upper': '╤', 'lower': '╧' }
        capo   = { 'upper': '╔', 'lower': '╚' }
        final  = { 'upper': '╕', 'lower': '╛' }
        fine   = { 'upper': '╗', 'lower': '╝' }
        total_fret_width = self.total_fret_width()
        render = ' ' * ( self.NOTE_WIDTH + self.string_n_size )
        render += capo[side]
        for f in range(1, frets * total_fret_width + 1):
            # final fret
            if f == frets * total_fret_width:
                if f % (total_fret_width * 12) == 0:
                    render += fine[side]
                else:
                    render += final[side]
            # 12th, 24th
            elif f % (total_fret_width * 12) == 0:
                render += twelve[side]
            # fret bar joints
            elif f % total_fret_width == 0:
                render += joints[side]
            # normal binding
            else:
                render += normal[side]
        echo(render)


    def render_string(self, s, mode, frets=12, verbose=1, show_degrees=False):
        string_n = FString(
            s,
            fg='cyan',
            fx=['faint' if verbose < 1 else ''],
            size=self.string_n_size,
        )
        string = self.strings[s-1]
        string.mode = mode
        string.frets = frets
        string.verbose = verbose
        string.show_degrees = show_degrees
        echo(string_n + string)


    def render_fretboard(self, mode=None, frets=12, verbose=1, show_degrees=False):
        # if frets > max_frets_on_screen():
        #     frets = max_frets_on_screen()
        self.render_inlays(frets, verbose)
        self.render_binding('upper', frets)
        for s, _ in enumerate(self.strings):
            self.render_string(s+1, mode, frets, verbose, show_degrees)
        self.render_binding('lower', frets)


def max_frets_on_screen():
    frets_allowed_by_tty = int(
        tty_cols() / PluckedStringInstrument.total_fret_width()
    ) - 2
    if frets_allowed_by_tty < PluckedStringInstrument.maximum_frets():
        return frets_allowed_by_tty
    return PluckedStringInstrument.maximum_frets()




from bestia.output import Row, FString

from ..notes import NotePitch
from ..notes.intervals import Intervals

from ..errors import InvalidNote, InvalidOctave

__all__ = [
    'TonalKey',
    'MajorScale',
    'MajorPentatonicScale',
    'AugmentedScale',
    'IonianMode',
    'MixolydianMode',
    'LydianMode',
    'MinorScale',
    'DiminishedScale',
    'MinorPentatonicScale',
    'MelodicMinorScale',
    'HarmonicMinorScale',
    'AeolianMode',
    'DorianMode',
    'PhrygianMode',
    'LocrianMode',
    'ChromaticScale',
]


class TonalKey:

    notations = ()
    intervals = ()
    degrees = ()

    @classmethod
    def _calc_intervals(cls):
        return cls.intervals


    @classmethod
    def allowed_degrees(cls):
        ''' returns list of each degree, once '''
        if cls.degrees:
            return list(cls.degrees)
        return [ n+1 for n in range( len(cls._calc_intervals()) ) ]


    @classmethod
    def all_degrees(cls):
        ''' returns list of each degree, over all octaves '''
        degrees = cls.allowed_degrees()

        # floors over-octave degrees into in-octave
        for d, deg in enumerate(degrees):
            if deg > len(cls._calc_intervals()):
                degrees[d] -= len(cls._calc_intervals())
                # ie.  9th  => 2nd
                # ie.  11th => 4th

        # remove duplicates
        degrees = list( set(degrees) )

        # sort order
        degrees.sort()

        # calculate all possible octaves
        all_degrees = []
        for o in range(NotePitch.MAXIMUM_OCTAVE):
            for deg in degrees:
                all_degrees.append(
                    deg + len(cls._calc_intervals()) * o
                )

        return tuple(all_degrees)


    @classmethod
    def name(cls):
        n = ''
        for c in cls.__name__:
            if c.isupper():
                n += ' '
            n += c.lower()
        return n.strip()


    @classmethod
    def __possible_root_notes(cls, valids=True):
        ''' checks all possible notes for validity as root of given key '''
        valid_roots = []
        invalid_roots = []

        for note in NotePitch.notes_by_alts():

            try:
                invalid_root = False
                for _ in cls(*note).spell(
                    note_count=len(cls._calc_intervals()) +1,
                    yield_all=False,
                ):
                    # if any degree fails, scale is not spellable
                    pass

            except InvalidNote:
                invalid_root = True

            finally:
                if invalid_root:
                    invalid_roots.append(note)
                else:
                    valid_roots.append(note)

        return valid_roots if valids else invalid_roots


    @classmethod
    def valid_root_notes(cls):
        ''' returns only valid root notes for given key class '''
        return cls.__possible_root_notes(valids=True)


    @classmethod
    def invalid_root_notes(cls):
        ''' returns only invalid root notes for given key class '''
        return cls.__possible_root_notes(valids=False)


    def __init__(self, chr, alt='', oct=0):
        self.root = NotePitch(chr, alt, 0) # ignore note.oct


    def __repr__(self):
        ''' prints first octave of NotePitch items '''
        spell_line = Row()
        for degree in self.spell(
            note_count=None, yield_all=False
        ):
            spell_line.append( FString(degree, size=5) )
        spell_line.append( f'{type(self)} on {self.root}'[:-1] )
        return str(spell_line)


    def __getitem__(self, d):

        if d < 1:
            return False

        if self.degrees and d not in self.all_degrees():
            return None

        if d == 1:
            return self.root

        # GET DEGREE's ROOT OFFSETS = PERFECT_OCTAVE + SPARE_STS
        octs_from_root, spare_sts = divmod(
            self.degree_root_interval(d), Intervals.PERFECT_OCTAVE
        )
        note_oct = octs_from_root

        # GET DEGREE PROPERTIES FROM ENHARMONIC MATRIX
        next_notes = [
            n for n in NotePitch.EnharmonicMatrix()[
                self.root.enharmonic_row + spare_sts
            ] if n.chr == self.root.adjacent_chr(d - 1) # EXPECTED TONE
        ]

        if len(next_notes) == 1:
            # AT THIS POINT DEG_OCT CAN EITHER STAY | +1
            if self.root.oversteps_oct(next_notes[0]):
                note_oct += 1

            # RETURN NEW OBJECT, DO NOT CHANGE ENHARMONIC MATRIX ITEM!
            return NotePitch(next_notes[0].chr, next_notes[0].alt, note_oct)

        raise InvalidNote


    def validate(self):
        for note in self.invalid_root_notes():
            if note ** self.root:
                return False
        return True


    def degree_root_interval(self, d):
        ''' return degree's delta semitones from key's root '''
        if d > len(self._calc_intervals()):
            return self.degree_root_interval(
                d - len(self._calc_intervals())
            ) + Intervals.PERFECT_OCTAVE
        return self._calc_intervals()[d -1]


    def _count_notes(self, note_count=-1, start_note=None, yield_all=True):
        ''' * returns when note_count == 0
            * positive note_count yields notes exactly
            * negative note_count yields notes indefinetly
            * filters received Nones when needed
        '''
        if not start_note:
            start_note = self.root

        for note in self.__solmizate(start_note=start_note):

            if note_count == 0:
                return

            if note:
                yield note
                note_count -= 1

            elif yield_all:
                yield None


    def __solmizate(self, start_note):
        ''' uses __get_item__ to retrieve notes to yield:
            before yielding a note, calulates semitone
            difference with previous note and yields Nones

            * yields based on all_degrees
            * always yields Nones
            * changes octs by using self[d]
            * start_note (diatonic, non-diatonic)
        '''
        degrees = self.all_degrees()

        for d, _ in enumerate(degrees):

            this = self[ degrees[d] ]

            if this < start_note:
                ## start_note not yet reached
                continue

            #####################################################
            # calculate Nones to yield for non-diatonic semitones
            #####################################################
            if this != start_note: # do not yield the Nones before starting note

                prev = self[ degrees[d-1] ]
                if prev >= start_note:
                    # dont yield last st, yield the note below
                    last_interval = this - prev - 1

                else:
                    # to calculate Nones to yield, dont go all way back to prev
                    # when prev is lower than starting note
                    last_interval = this - start_note

                for semitone in range(last_interval):
                    yield None

            yield this


    def spell(self, note_count=None, start_note=None, yield_all=False):
        if note_count == None:
            note_count = len(
                self.degrees if self.degrees else self._calc_intervals()
            ) + 1  # also include next octave root note
        try:
            return self._count_notes(
                note_count=note_count,
                start_note=start_note,
                yield_all=yield_all,
            )
        except InvalidOctave:
            pass


########################
### MAJOR KEYS/MODES ###
########################

class MajorScale(TonalKey):
    notations = (
        'major',
    )
    intervals = (
        Intervals.UNISON,
        Intervals.MAJOR_SECOND,
        Intervals.MAJOR_THIRD,
        Intervals.PERFECT_FOURTH,
        Intervals.PERFECT_FIFTH,
        Intervals.MAJOR_SIXTH,
        Intervals.MAJOR_SEVENTH,
    )

class MajorPentatonicScale(MajorScale):
    notations = (
        'major_pentatonic',
    )
    degrees = (1, 2, 3, 5, 6)

class IonianMode(MajorScale):
    notations = (
        'ionian',
    )

class MixolydianMode(MajorScale):
    notations = (
        'mixolydian',
    )
    intervals = (
        Intervals.UNISON,
        Intervals.MAJOR_SECOND,
        Intervals.MAJOR_THIRD,
        Intervals.PERFECT_FOURTH,
        Intervals.PERFECT_FIFTH,
        Intervals.MAJOR_SIXTH,
        Intervals.MINOR_SEVENTH, # <<<
    )

class LydianMode(MajorScale):
    notations = (
        'lydian',
    )
    intervals = (
        Intervals.UNISON,
        Intervals.MAJOR_SECOND,
        Intervals.MAJOR_THIRD,
        Intervals.AUGMENTED_FOURTH, # <<<
        Intervals.PERFECT_FIFTH,
        Intervals.MAJOR_SIXTH,
        Intervals.MAJOR_SEVENTH,
    )

########################
### MINOR KEYS/MODES ###
########################

class MinorScale(TonalKey):
    notations = (
        'minor',
    )
    intervals = (
        Intervals.UNISON,
        Intervals.MAJOR_SECOND,
        Intervals.MINOR_THIRD,
        Intervals.PERFECT_FOURTH,
        Intervals.PERFECT_FIFTH,
        Intervals.MINOR_SIXTH,
        Intervals.MINOR_SEVENTH,
    )

class MinorPentatonicScale(MinorScale):
    notations = (
        'minor_pentatonic',
    )
    degrees = (1, 3, 4, 5, 7)

class MelodicMinorScale(MinorScale):
    notations = (
        'melodic_minor',
    )
    intervals = (
        Intervals.UNISON,
        Intervals.MAJOR_SECOND,
        Intervals.MINOR_THIRD,
        Intervals.PERFECT_FOURTH,
        Intervals.PERFECT_FIFTH,
        Intervals.MAJOR_SIXTH, # <<<
        Intervals.MAJOR_SEVENTH, # <<<
    )

class HarmonicMinorScale(MinorScale):
    notations = (
        'harmonic_minor',
    )
    intervals = (
        Intervals.UNISON,
        Intervals.MAJOR_SECOND,
        Intervals.MINOR_THIRD,
        Intervals.PERFECT_FOURTH,
        Intervals.PERFECT_FIFTH,
        Intervals.MINOR_SIXTH,
        Intervals.MAJOR_SEVENTH, # <<<
    )

class AeolianMode(MinorScale):
    notations = (
        'aeolian',
    )

class DorianMode(MinorScale):
    notations = (
        'dorian',
    )
    intervals = (
        Intervals.UNISON,
        Intervals.MAJOR_SECOND,
        Intervals.MINOR_THIRD,
        Intervals.PERFECT_FOURTH,
        Intervals.PERFECT_FIFTH,
        Intervals.MAJOR_SIXTH, # <<<
        Intervals.MINOR_SEVENTH,
    )

class PhrygianMode(MinorScale):
    notations = (
        'phrygian',
    )
    intervals = (
        Intervals.UNISON,
        Intervals.MINOR_SECOND, # <<<
        Intervals.MINOR_THIRD,
        Intervals.PERFECT_FOURTH,
        Intervals.PERFECT_FIFTH,
        Intervals.MINOR_SIXTH,
        Intervals.MINOR_SEVENTH,
    )

class LocrianMode(MinorScale):
    notations = (
        'locrian',
    )
    intervals = (
        Intervals.UNISON,
        Intervals.MINOR_SECOND, # <<<
        Intervals.MINOR_THIRD,
        Intervals.PERFECT_FOURTH,
        Intervals.DIMINISHED_FIFTH, # <<<
        Intervals.MINOR_SIXTH,
        Intervals.MINOR_SEVENTH,
    )


#####################
### CHROMATIC KEY ###
#####################

class ChromaticScale(TonalKey):
    notations = (
        'chromatic',
    )
    intervals = (
        Intervals.UNISON,
        Intervals.MINOR_SECOND,
        Intervals.MAJOR_SECOND,
        Intervals.MINOR_THIRD,
        Intervals.MAJOR_THIRD,
        Intervals.PERFECT_FOURTH,
        Intervals.AUGMENTED_FOURTH,
        Intervals.PERFECT_FIFTH,
        Intervals.MINOR_SIXTH,
        Intervals.MAJOR_SIXTH,
        Intervals.MINOR_SEVENTH,
        Intervals.MAJOR_SEVENTH,
    )

    def __getitem__(self, d):

        if d < 1:
            return False

        if d == 1:
            return self.root

        # GET DEGREE's ROOT OFFSETS = PERFECT_OCTAVE + SPARE_STS
        octs_from_root, spare_sts = divmod(
            self.degree_root_interval(d), Intervals.PERFECT_OCTAVE
        )
        note_oct = octs_from_root

        # GET DEGREE PROPERTIES FROM ENHARMONIC MATRIX
        # DO I REALLY NEED THESE 3 CHECKS ?
        # MATCH ROOT_TONE
        next_notes = [
            n for n in NotePitch.EnharmonicMatrix()[
                self.root.enharmonic_row + spare_sts
            ] if n ** self.root
        ]

        if not next_notes:
            # MATCH ROOT_ALT
            next_notes = [
                n for n in NotePitch.EnharmonicMatrix()[
                    self.root.enharmonic_row + spare_sts
                ] if n.alt == self.root.alt[:-1]
            ]

            if not next_notes:
                # CHOOSE "#" or ""
                chosen_alt = '#' if self.root.alt == '' else self.root.alt
                next_notes = [
                    n for n in NotePitch.EnharmonicMatrix()[
                        self.root.enharmonic_row + spare_sts
                    ] if n.alt == chosen_alt
                ]

        if len(next_notes) == 1:
            # AT THIS POINT DEG_OCT CAN EITHER STAY | +1
            if self.root.oversteps_oct(next_notes[0]):
                note_oct += 1

            # RETURN NEW OBJECT, DO NOT CHANGE ENHARMONIC MATRIX ITEM!
            return NotePitch(next_notes[0].chr, next_notes[0].alt, note_oct)

        raise InvalidNote


class DiminishedScale(ChromaticScale):
    notations = (
        'diminished',
    )
    degrees = (1, 3, 4, 6, 7, 9, 10, 12)

# TODO: decide which AugmentedScale definition works best

class AugmentedScale(ChromaticScale):
    ''' chromatic scale as parent enables augmented scales for __any__ root
    '''
    notations = (
        'augmented',
    )
    degrees = (1, 4, 5, 8, 9, 12)

# class AugmentedScale(TonalKey):
#     ''' tonal scale as parent allows for combination of flats and sharps
#     '''
#     notations = (
#         'augmented',
#     )
#     intervals = (
#         Intervals.UNISON,  # C
#         Intervals.AUGMENTED_SECOND,  # D#
#         Intervals.MAJOR_THIRD,  # E
#         Intervals.PERFECT_FOURTH, # F  exclude this interval using degrees...
#         Intervals.PERFECT_FIFTH,  # G
#         Intervals.MINOR_SIXTH,  # Ab
#         Intervals.MAJOR_SEVENTH,  # B
#     )
#     degrees = (1, 2, 3,  5, 6, 7)



'''
degrees tuple is not explicitly defined anymore but is inherited from parent class:

    class TriadChord(Chord):
    class SeventhChord(Chord):
    class SixthChord(Chord):
    class NinthChord(Chord):


parent scale is now a separate class atrribute
parent_scale_degree allows to form chords from degrees other than I


'''

from .scales import (
    MajorScale, MinorScale, AugmentedScale, DiminishedScale,
    IonianMode, AeolianMode, MixolydianMode, LocrianMode, DorianMode,
    HarmonicMinorScale, MelodicMinorScale
)
from .scales import TonalKey

from ..notes.intervals import Intervals

from ..errors import InvalidNote


__all__ = [
    'PowerChord',
    'Suspended4Chord',
    'Suspended2Chord',

    'MajorTriad',
    'MinorTriad',
    'AugmentedTriad',
    'DiminishedTriad',

    'MajorSixthChord',
    'MinorSixthChord',

    'MajorSeventhChord',
    'MinorSeventhChord',
    'DominantSeventhChord',
    'HalfDiminishedSeventhChord',
    'DiminishedSeventhChord',

    'MajorAdd9Chord',
    'MinorAdd9Chord',
    'AugmentedAdd9Chord',
    'DiminishedAdd9Chord',

    'MajorNinthChord',
    'MinorNinthChord',
    'DominantNinthChord',
    'DominantMinorNinthChord',
]


class Chord(TonalKey):
    parent_scale = MajorScale
    parent_scale_degree = 1
    degrees = (1, 3, 5,)

    @classmethod
    def find_parent_scale_root(cls, match_note):
        ''' finds char for which your parent_scale will match your chord's root
            at parent_scale_degree
        '''
        scales_found = []
        for n in cls.parent_scale.valid_root_notes():
            scale = cls.parent_scale(*n)
            if scale[cls.parent_scale_degree] ** match_note:
                scales_found.append(scale)

        assert len(scales_found) <= 1
        if scales_found:
            return scales_found[0]
        return None


    @classmethod
    def _parent_scale_root_offset(cls) -> int:
        ''' if I am an E7 chord and my parent_scale is A Major
            my offset from parent_scale root is 7 semitones
        '''
        return cls.parent_scale._calc_intervals()[cls.parent_scale_degree - 1]


    @classmethod
    def _calc_intervals(cls):
        # intervals is an empty tuple for chords
        offset_from_root = cls._parent_scale_root_offset()
        arranged_intervals = []
        for parent_scale_interval in cls.parent_scale._calc_intervals():
            new_interval = parent_scale_interval - offset_from_root
            if new_interval < 0:
                new_interval += Intervals.PERFECT_OCTAVE
            arranged_intervals.append(new_interval)
        arranged_intervals.sort()
        return arranged_intervals


class PowerChord(Chord):
    ''' C  G  '''
    notations = (
        '5',
    )
    degrees = (1, 5, )

class SusFourChord(Chord):
    degrees = (1, 4, 5, )

class SusTwoChord(Chord):
    degrees = (1, 2, 5, )

class TriadChord(Chord):
    degrees = (1, 3, 5, )

class SixthChord(Chord):
    degrees = (1, 3, 5, 6, )

class SeventhChord(Chord):
    degrees = (1, 3, 5, 7, )

class AddNineChord(Chord):
    degrees = (1, 3, 5, 9, )

class NinthChord(Chord):
    degrees = (1, 3, 5, 7, 9, )


########################
### SUSPENDED CHORDS ###
########################

class Suspended4Chord(SusFourChord):
    ''' C  F  G '''
    parent_scale = IonianMode
    notations = (
        'sus4',
        'sus',
    )

class Suspended2Chord(SusTwoChord):
    ''' C  D  G '''
    parent_scale = IonianMode
    notations = (
        'sus2',
        'sus9',
    )


####################
### TRIAD CHORDS ###
####################

class MajorTriad(TriadChord):
    ''' C  E  G  '''
    parent_scale = MajorScale
    notations = (
        'maj',
        '',
        'major',
    )

class MinorTriad(TriadChord):
    ''' C  Eb G  '''
    parent_scale = MinorScale
    notations = (
        'min',
        '-',
        'm',
        'minor',
    )

class AugmentedTriad(TriadChord):
    ''' C  E  G#  '''
    parent_scale = HarmonicMinorScale
    parent_scale_degree = 3
    notations = (
        'aug',
        'augmented',
    )

class DiminishedTriad(TriadChord):
    ''' C  Eb Gb  '''
    parent_scale = MajorScale
    parent_scale_degree = 7
    notations = (
        'dim',
        'diminished',
    )


####################
### SIXTH CHORDS ###
####################

class MajorSixthChord(SixthChord):
    ''' C  E  G  A '''
    parent_scale = IonianMode
    notations = (
        '6',
        'add6',
    )

class MinorSixthChord(SixthChord):
    ''' C  Eb G  A '''
    parent_scale = DorianMode
    notations = (
        'm6',
        # 'madd6',
        'min6',
    )


######################
### SEVENTH CHORDS ###
######################

class MajorSeventhChord(SeventhChord):
    ''' C  E  G  B  '''
    parent_scale = IonianMode
    notations = (
        'maj7',
        'M7',  # careful with these 2 if ever using .lower() to compare
        'Δ7',
        'Δ',
        'major7',
    )

class MinorSeventhChord(SeventhChord):
    ''' C  Eb G  Bb '''
    parent_scale = AeolianMode
    notations = (
        'min7',
        'm7',  # careful with these 2 if ever using .lower() to compare
        '-7',
        'minor7',
    )

class DominantSeventhChord(SeventhChord):
    ''' C  E  G  Bb '''
    parent_scale = MixolydianMode
    notations = (
        '7',
        'dom7',
        'dominant7',
    )

class HalfDiminishedSeventhChord(SeventhChord):
    ''' C  Eb Gb Bb '''
    parent_scale = LocrianMode
    notations = (
        'm7b5',
        'm7-5',
        'min7dim5',
        'm7(b5)',
        'ø7',
    )

class DiminishedSeventhChord(SeventhChord):
    ''' C  Eb Gb Bbb'''
    parent_scale = HarmonicMinorScale
    parent_scale_degree = 7
    notations = (
        'dim7',
        'o7',
        'diminished7',
    )


###################
### ADD9 CHORDS ###
###################

class MajorAdd9Chord(AddNineChord):
    ''' C  E  G  D '''
    parent_scale = IonianMode
    notations = (
        'add9',
        'Add9',
    )

class MinorAdd9Chord(AddNineChord):
    ''' C  Eb G  D '''
    parent_scale = AeolianMode
    notations = (
        'madd9',
        'mAdd9',
    )

class AugmentedAdd9Chord(AddNineChord):
    ''' C  E  G# D '''
    parent_scale = HarmonicMinorScale
    parent_scale_degree = 3
    notations = (
        'augadd9',
        'augAdd9',
    )

class DiminishedAdd9Chord(AddNineChord):
    ''' C  Eb Gb D '''
    parent_scale = MelodicMinorScale
    parent_scale_degree = 6
    notations = (
        'dimadd9',
        'dimAdd9',
    )


####################
### NINTH CHORDS ###
####################

class MajorNinthChord(NinthChord):
    ''' C  E  G  B  D  '''
    parent_scale = IonianMode
    notations = (
        'maj9',
        'M9',
        'major9',
    )

class MinorNinthChord(NinthChord):
    ''' C  Eb G  Bb D  '''
    parent_scale = AeolianMode
    notations = (
        'm9',
        'min9',
        '-9',
        'minor9',
    )

class DominantNinthChord(NinthChord):
    ''' C  E  G  Bb D  '''
    parent_scale = MixolydianMode
    notations = (
        '9',
        'dom9',
    )

class DominantMinorNinthChord(NinthChord):
    ''' C  E  G  Bb Db '''
    parent_scale = HarmonicMinorScale
    parent_scale_degree = 5
    notations = (
        '7b9',
    )




from bestia.iterate import LoopedList

from .intervals import Intervals

from ..errors import InvalidNote, InvalidAlteration, InvalidOctave

__all__ = [
    'NotePitch',
]


class NotePitch:

    _CHARS = LoopedList(
        'C',
        None,
        'D',
        None,
        'E',
        'F',
        None,
        'G',
        None,
        'A',
        None,
        'B',
    )

    _ALTS = {
        'bb': '𝄫',
        'b': '♭',
        # '♮': '',
        '': '',
        '#': '♯',
        '##': '𝄪',
    }

    _OCTAVES = (
        # '₀',
        # '₁',
        # '₂',
        # '₃',
        # '₄',
        # '₅',
        # '₆',
        # '₇',
        # '₈',
        # '₉',
        '⁰',
        '¹',
        '²',
        '³',
        '⁴',
        '⁵',
        '⁶',
        '⁷',
        '⁸',
        '⁹',
    )

    DEFAULT_OCTAVE = 3
    MAXIMUM_OCTAVE = 9

    @classmethod
    def EnharmonicMatrix(cls):
        ''' This is the heart of the whole project
            indeces are used to determine:
            * when to change octs
            * intervals between NotePitch instances
            notes MUST be unique so that TonalKey[d] finds 1 exact match!
        '''
        return LoopedList(
            ## 2-octave enharmonic relationships
            (  cls('C', '' , 2), cls('B', '#' , 1), cls('D', 'bb', 2)  ), # NAH
            (  cls('C', '#', 2), cls('D', 'b' , 2), cls('B', '##', 1)  ), # AAH

            ## 1-octave enharmonic relationships
            (  cls('D', '' , 1), cls('C', '##', 1), cls('E', 'bb', 1)  ), # NHH
            (  cls('D', '#', 1), cls('E', 'b' , 1), cls('F', 'bb', 1)  ), # AAH
            (  cls('E', '' , 1), cls('F', 'b' , 1), cls('D', '##', 1)  ), # NAH
            (  cls('F', '' , 1), cls('E', '#' , 1), cls('G', 'bb', 1)  ), # NAH
            (  cls('F', '#', 1), cls('G', 'b' , 1), cls('E', '##', 1)  ), # AAH
            (  cls('G', '' , 1), cls('F', '##', 1), cls('A', 'bb', 1)  ), # NHH
            (  cls('G', '#', 1), cls('A', 'b' , 1)                     ), # AA
            (  cls('A', '' , 1), cls('G', '##', 1), cls('B', 'bb', 1)  ), # NHH

            ## 2-octave enharmonic relationships
            (  cls('A', '#', 1), cls('B', 'b' , 1), cls('C', 'bb', 2)  ), # AAH
            (  cls('B', '' , 1), cls('C', 'b' , 2), cls('A', '##', 1)  ), # NAH
        )

    @classmethod
    def possible_chars(cls):
        return [ c for c in cls._CHARS if c ]

    @classmethod
    def validate_char(cls, char):
        char = char.upper()
        if char not in cls.possible_chars():
            raise InvalidNote(char)
        return char

    @classmethod
    def input_alterations(cls) -> tuple:
        return tuple( cls._ALTS.keys() )

    @classmethod
    def output_alterations(cls) -> tuple:
        return tuple( cls._ALTS.values() )

    @classmethod
    def notes_by_alts(cls):
        ''' yields all 35 possible notes in following order:
            * 7 notes with alt ''
            * 7 notes with alt 'b'
            * 7 notes with alt '#'
            * 7 notes with alt 'bb'
            * 7 notes with alt '##'
        '''
        # sort alts
        alts = list( cls.input_alterations() )
        alts.sort(key=len) # '', b, #, bb, ##

        # get all notes
        notes = []
        for ehns in cls.EnharmonicMatrix():
            for ehn in ehns:
                notes.append(ehn)

        # yield notes
        for alt in alts:
            for note in notes:
                if note.alt == alt:
                    # note.oct = 3
                    yield note


    def __init__(self, char, *args):
        ''' init WITHOUT specifying argument names
            should be able to handle:
            NotePitch('C')
            NotePitch('C', 9)
            NotePitch('C', '#')
            NotePitch('C', '#', 9)
        '''
        self.chr = self.validate_char(char)
        self.alt = ''
        self.oct = self.DEFAULT_OCTAVE

        if len(args) > 2:
            raise InvalidNote('Too many arguments')

        # with only 1 arg, decide if it's alt or oct
        if len(args) == 1:
            if args[0] in self._ALTS:
                self.alt = args[0]
            else:
                try:
                    self.oct = int(args[0])
                except:
                    raise InvalidNote(
                        'Failed to parse argument: '.format(args[0])
                    )

        # with 2 args, order must be alt, oct
        if len(args) == 2:
            if args[0] not in self._ALTS:
                raise InvalidAlteration(args[0])
            self.alt = args[0]

            try:
                self.oct = int(args[1])
            except:
                raise InvalidOctave(args[1])

        if self.oct > self.MAXIMUM_OCTAVE:
            raise InvalidOctave(self.oct)


    def __iter__(self):
        ''' allows unpacking Note objects as args for other functions '''
        for i in (self.chr, self.alt, self.oct):
            yield i

    def __repr__(self):
        return '{}{}{}'.format(
            self.chr,
            self.repr_alt,
            self.repr_oct,
        )

    @property
    def repr_oct(self):
        output = ''
        for c in str(self.oct):
            output += self._OCTAVES[int(c)]
        return output

    @property
    def repr_alt(self):
        return self._ALTS[self.alt]

    def __sub__(self, other):
        if self.__class__ == other.__class__:
            oct_interval = (self.oct - other.oct) * Intervals.PERFECT_OCTAVE
            chr_interval = self._CHARS.index(self.chr) - self._CHARS.index(other.chr)
            alt_interval = (
                self.input_alterations().index(self.alt) - self.input_alterations().index(other.alt)
            )
            return oct_interval + chr_interval + alt_interval
        raise TypeError(' - not supported with {}'.format(other.__class__))

    def __pow__(self, other):
        if self.__class__ == other.__class__:
            if self.chr == other.chr:
                if self.alt == other.alt:
                    return True
            return False
        raise TypeError(' ** not supported with {}'.format(other.__class__))

    def __rshift__(self, other):
        if self.__class__ == other.__class__:
            if self.chr == other.chr:
                if self.alt == other.alt:
                    if self.oct == other.oct:
                        return True
            return False
        raise TypeError(' >> not supported with {}'.format(other.__class__))

    def __eq__(self, other):
        if other.__class__ == self.__class__:
            return other - self == Intervals.UNISON

    def __ne__(self, other):
        if other.__class__ == self.__class__:
            return other - self != Intervals.UNISON
        return True

    def __gt__(self, other):
        if self.__class__ == other.__class__:
            return self - other > Intervals.UNISON
        raise TypeError(' > not supported with {}'.format(other.__class__))

    def __ge__(self, other):
        if self.__class__ == other.__class__:
            return self - other >= Intervals.UNISON
        raise TypeError(' >= not supported with {}'.format(other.__class__))

    def __lt__(self, other):
        if self.__class__ == other.__class__:
            return self - other < Intervals.UNISON
        raise TypeError(' < not supported with {}'.format(other.__class__))

    def __le__(self, other):
        if self.__class__ == other.__class__:
            return self - other <= Intervals.UNISON
        raise TypeError(' <= not supported with {}'.format(other.__class__))


    ### CHR METHODS
    def relative_chr(self, n):
        my_index = self._CHARS.index(self.chr)
        return self._CHARS[my_index +n]

    def adjacent_chr(self, n=1):
        ''' returns next adjacent chr of self
            negative fails...
        '''
        c = ''
        i = 1
        while n != 0:
            c = self.relative_chr(i)
            if c:
                n -= 1
            i += 1
        return c

    ### OCT METHODS
    def oversteps_oct(self, other):
        ''' evaluates whether I need to cross octave border in order to go from self to other
            ignores octave
        '''
        if self.chr != other.chr:
            return self._CHARS.index(self.chr) > self._CHARS.index(other.chr)
        return (
            self.input_alterations().index(self.alt) > self.input_alterations().index(other.alt)
        )


    # ENHARMONIC ATTRIBUTES
    @property
    def enharmonic_row(self):
        return self.matrix_coordinates[0]

    @property
    def matrix_coordinates(self):
        for row_index, row in enumerate( self.EnharmonicMatrix() ):
            for note_index, enharmonic_note in enumerate(row):
                if self ** enharmonic_note:  # ignore oct
                    return (row_index, note_index)




"""
this module is in charge of loading the data of the .json files in the tunings directory and make it available to the fretboard application.
"""

import os
import json

from kord.notes import NotePitch

JSON_DIR = '{}/tunings'.format( os.path.dirname(os.path.realpath(__file__)) )

def json_instruments():
    return [
        f.split('.')[0] for f in os.listdir(JSON_DIR) if f.endswith('.json')
    ]

def open_instrument(instrument):
    try:
        with open('{}/{}.json'.format(JSON_DIR, instrument)) as js:
            return json.load(js)
    except:
        return {}

def load_tuning_data():
    data = {}
    for instrument in json_instruments():

        instrument_data = open_instrument(instrument)
        if not instrument_data:
            print('WARNING: {}.json is not valid json, ignoring file...'.format(instrument))
            continue

        for tuning, strings_list in instrument_data.items():
            for s, string in enumerate(strings_list):
                try:
                    # validates the note attributes
                    instrument_data[tuning][s] = NotePitch(
                        string[0], string[1:-1], string[-1]
                    )
                except:
                    print(
                        'WARNING: "{}" is not a valid note, ignoring {}.json "{}" tuning...'.format(string, instrument, tuning)
                    )

        data[instrument] = instrument_data

    return data




from ..notes import NotePitch
from ..keys.chords import *

from ..errors import InvalidNote, InvalidAlteration, InvalidOctave, InvalidChord

from bestia.output import echo

__all__ = [
    'NotePitchParser',
]


class NotePitchParser:

    def __init__(self, symbol):
        self.symbol = symbol
        self.reset()


    def reset(self):
        self.to_parse = self.symbol.strip()


    def _parse_char(self):
        char = NotePitch.validate_char(self.to_parse[0])
        self.to_parse = self.to_parse[1:]
        return char


    def _parse_oct(self):
        if len(self.to_parse) > 1:
            # use str.isdigit()
            try:
                # this being an int means octave > MAXIMUM_OCTAVE
                octave = int(self.to_parse[-2])
                octave_over_max = True
            except ValueError:
                # probably an alt
                octave_over_max = False
            finally:
                if octave_over_max:
                    raise InvalidOctave(self.symbol)

        try:
            # this being an int means this is an octave
            octave = int(self.to_parse[-1])
            self.to_parse = self.to_parse[:-1]
        except ValueError:
            # probably an alt
            octave = NotePitch.DEFAULT_OCTAVE
        finally:
            return octave


    def _parse_alts(self):
        self.to_parse = self.to_parse.replace('𝄫', 'bb')
        self.to_parse = self.to_parse.replace('♭', 'b')
        self.to_parse = self.to_parse.replace('𝄪', '##')
        self.to_parse = self.to_parse.replace('♯', '#')

        if self.to_parse not in NotePitch.input_alterations():
            raise InvalidAlteration(self.symbol)

        alts = self.to_parse
        self.to_parse = self.to_parse[len(self.to_parse):]
        return alts


    def parse(self):
        if len(self.to_parse) == 0:
            raise InvalidNote(self.symbol)

        char = self._parse_char()
        if len(self.to_parse) == 0:
            self.reset()
            return NotePitch(char)

        octave = self._parse_oct()
        if len(self.to_parse) == 0:
            self.reset()
            return NotePitch(char, octave)

        alts = self._parse_alts()
        if len(self.to_parse) == 0:
            self.reset()
            return NotePitch(char, alts, octave)

        raise InvalidNote(self.symbol)




import unittest
import random

from bestia.output import echo, Row, FString

from .scales import (
    ChromaticScale, MajorPentatonicScale, MajorScale,
    MinorScale, MelodicMinorScale, HarmonicMinorScale
)

from ..notes import NotePitch
from ..notes.constants import *

from ..errors import InvalidOctave

__all__ = [
    'ScaleValidityTest',
    'ChromaticScalesTest',
    'MajorScalesExpectedNotesTest',
    'TonalScaleSpellMethodTest',
]


class ScaleValidityTest(unittest.TestCase):

    ''' for a given key, allows to verify:
            * invalid roots
            * octave changes
    '''

    def setUp(self):
        print()

        self.chromatic_key = ChromaticScale
        # test no invalid roots

        self.major_key = MajorScale
        # F𝄫¹ invalid MajorScale
        # B𝄪¹ invalid MajorScale
        # D𝄪¹ invalid MajorScale
        # E𝄪¹ invalid MajorScale
        # G𝄪¹ invalid MajorScale
        # A𝄪¹ invalid MajorScale

        self.minor_key = MinorScale
        # D𝄫² invalid MinorScale
        # F𝄫¹ invalid MinorScale
        # G𝄫¹ invalid MinorScale
        # C𝄫² invalid MinorScale
        # B𝄪¹ invalid MinorScale
        # E𝄪¹ invalid MinorScale

        self.mel_minor_key = MelodicMinorScale
        # F𝄫¹ invalid MelodicMinorScale
        # G𝄫¹ invalid MelodicMinorScale
        # C𝄫² invalid MelodicMinorScale
        # B𝄪¹ invalid MelodicMinorScale
        # D𝄪¹ invalid MelodicMinorScale
        # E𝄪¹ invalid MelodicMinorScale
        # G𝄪¹ invalid MelodicMinorScale
        # A𝄪¹ invalid MelodicMinorScale

        self.har_minor_key = HarmonicMinorScale
        # D𝄫² invalid HarmonicMinorScale
        # F𝄫¹ invalid HarmonicMinorScale
        # G𝄫¹ invalid HarmonicMinorScale
        # C𝄫² invalid HarmonicMinorScale
        # B𝄪¹ invalid HarmonicMinorScale
        # D𝄪¹ invalid HarmonicMinorScale
        # E𝄪¹ invalid HarmonicMinorScale
        # G𝄪¹ invalid HarmonicMinorScale
        # A𝄪¹ invalid HarmonicMinorScale


    def testValidMethod(self):
        for Scale in [self.major_key]:
            assert Scale('c').validate()


    def testValidRoots(self):

        for Scale in [self.major_key,self.minor_key ]:

            echo('\nValid {}s'.format(Scale.__name__), 'underline')

            for note in Scale.valid_root_notes():

                line = Row()
                key = Scale(*note)
                for d in key.spell(
                    note_count=len(key.intervals) +16, yield_all=False
                ):
                    line.append(
                        FString(
                            d,
                            size=5,
                            fg='blue' if not (d.oct % 2) else 'white',
                        )
                    )
                line.echo()


    def testInvalidRoots(self):
        echo('\nInvalid {}s'.format(self.major_key.__name__), 'underline')
        for note in self.major_key.invalid_root_notes():
            echo(
                '{}{} invalid {}'.format(note.chr, note.repr_alt, self.major_key.__name__),
                'red', 'faint'
            )


class ChromaticScalesTest(unittest.TestCase):

    def setUp(self):
        print()
        self.c_chromatic = ChromaticScale('C')
        self.f_sharp_chromatic = ChromaticScale('F', '#')
        self.b_flat_chromatic = ChromaticScale('B', 'b')


    def testIntervalsCount(self):
        assert len(self.c_chromatic.intervals) == 12, self.c_chromatic.intervals
        assert len(self.f_sharp_chromatic.intervals) == 12, self.f_sharp_chromatic.intervals
        assert len(self.b_flat_chromatic.intervals) == 12, self.b_flat_chromatic.intervals


    def testCChromaticScaleGenerator(self):

        octaves_to_test = 18
        intervals = 12
        notes_to_test = octaves_to_test * intervals + 1  # 18 * 12 + 1 = 217

        for i, note in enumerate(
            self.c_chromatic.spell(
                note_count=notes_to_test, yield_all=False
            )
        ):
            i += 1
            print(i)
            if i == 1:
                assert note >> C_0, note
            elif i == 2:
                assert note >> C_SHARP_0, note
            elif i == 3:
                assert note >> D_0, note
            elif i == 4:
                assert note >> D_SHARP_0, note
            elif i == 5:
                assert note >> E_0, note
            elif i == 6:
                assert note >> F_0, note
            elif i == 7:
                assert note >> F_SHARP_0, note
            elif i == 8:
                assert note >> G_0, note
            elif i == 9:
                assert note >> G_SHARP_0, note
            elif i == 10:
                assert note >> A_0, note
            elif i == 11:
                assert note >> A_SHARP_0, note
            elif i == 12:
                assert note >> B_0, note
            elif i == 13:
                assert note >> C_1, note
            elif i == 14:
                assert note >> C_SHARP_1, note
            elif i == 15:
                assert note >> D_1, note
            elif i == 16:
                assert note >> D_SHARP_1, note
            elif i == 17:
                assert note >> E_1, note
            elif i == 18:
                assert note >> F_1, note
            elif i == 19:
                assert note >> F_SHARP_1, note
            elif i == 20:
                assert note >> G_1, note
            elif i == 21:
                assert note >> G_SHARP_1, note
            elif i == 22:
                assert note >> A_1, note
            elif i == 23:
                assert note >> A_SHARP_1, note
            elif i == 24:
                assert note >> B_1, note
            elif i == 25:
                assert note >> C_2, note
            # ..............................
            elif i == 97:
                assert note >> C_8, note
            elif i == 98:
                assert note >> C_SHARP_8, note
            elif i == 99:
                assert note >> D_8, note
            elif i == 100:
                assert note >> D_SHARP_8, note
            elif i == 101:
                assert note >> E_8, note
            elif i == 102:
                assert note >> F_8, note
            elif i == 103:
                assert note >> F_SHARP_8, note
            elif i == 104:
                assert note >> G_8, note
            elif i == 105:
                assert note >> G_SHARP_8, note
            elif i == 106:
                assert note >> A_8, note
            elif i == 107:
                assert note >> A_SHARP_8, note
            elif i == 108:
                assert note >> B_8, note
            elif i == 109:
                assert note >> C_9, note
            # ..............................
            elif i == 205:
                assert note >> NotePitch('C', '', 17), note


    def testFSharpChromaticScaleGenerator(self):

        octaves_to_test = 18
        intervals = 12
        notes_to_test = octaves_to_test * intervals + 1  # 18 * 12 + 1 = 217

        for i, note in enumerate(
            self.f_sharp_chromatic.spell(
                note_count=notes_to_test, yield_all=False
            )
        ):
            i += 1
            if i == 1:
                assert note >> F_SHARP_0, note
            elif i == 2:
                assert note >> G_0, note
            elif i == 3:
                assert note >> G_SHARP_0, note
            elif i == 4:
                assert note >> A_0, note
            elif i == 5:
                assert note >> A_SHARP_0, note
            elif i == 6:
                assert note >> B_0, note
            elif i == 7:
                assert note >> C_1, note
            elif i == 8:
                assert note >> C_SHARP_1, note
            elif i == 9:
                assert note >> D_1, note
            elif i == 10:
                assert note >> D_SHARP_1, note
            elif i == 11:
                assert note >> E_1, note
            elif i == 12:
                assert note >> F_1, note
            elif i == 13:
                assert note >> F_SHARP_1, note
            elif i == 14:
                assert note >> G_1, note
            elif i == 15:
                assert note >> G_SHARP_1, note
            elif i == 16:
                assert note >> A_1, note
            elif i == 17:
                assert note >> A_SHARP_1, note
            elif i == 18:
                assert note >> B_1, note
            elif i == 19:
                assert note >> C_2, note
            # ..............................
            elif i == 211:
                assert note >> NotePitch('C', '', 18), note


    def testBFlatChromaticScaleGenerator(self):

        octaves_to_test = 18
        intervals = 12
        notes_to_test = octaves_to_test * intervals + 1  # 18 * 12 + 1 = 217

        for i, note in enumerate(
            self.b_flat_chromatic.spell(
                note_count=notes_to_test, yield_all=False
            )
        ):
            i += 1
            if i == 1:
                assert note >> B_FLAT_0, note
            elif i == 2:
                assert note >> B_0, note
            elif i == 3:
                assert note >> C_1, note
            elif i == 4:
                assert note >> D_FLAT_1, note
            elif i == 5:
                assert note >> D_1, note
            elif i == 6:
                assert note >> E_FLAT_1, note
            elif i == 7:
                assert note >> E_1, note
            elif i == 8:
                assert note >> F_1, note
            elif i == 9:
                assert note >> G_FLAT_1, note
            elif i == 10:
                assert note >> G_1, note
            elif i == 11:
                assert note >> A_FLAT_1, note
            elif i == 12:
                assert note >> A_1, note
            elif i == 13:
                assert note >> B_FLAT_1, note
            elif i == 14:
                assert note >> B_1, note
            elif i == 15:
                assert note >> C_2, note
            # ..............................
            elif i == 200:
                assert note >> NotePitch('F', '', 17), note


class MajorScalesExpectedNotesTest(unittest.TestCase):

    def setUp(self):
        print()
        self.c_major = MajorScale('C')
        self.b_major = MajorScale('B')            # 5 sharps
        self.d_flat_major = MajorScale('D', 'b')  # 5 flats

    def testIntervalsCount(self):
        assert len(self.c_major.intervals) == 7, self.c_major.intervals
        assert len(self.b_major.intervals) == 7, self.b_major.intervals
        assert len(self.d_flat_major.intervals) == 7, self.d_flat_major.intervals


    def testCMajorScaleGenerator(self):

        octaves_to_test = 18
        intervals = 7
        notes_to_test = octaves_to_test * intervals + 1  # 18 * 7 + 1 = 127

        for i, note in enumerate(
            self.c_major.spell(
                note_count=notes_to_test, yield_all=False
            )
        ):
            i += 1
            if i == 1:
                assert note >> C_0, note
            elif i == 2:
                assert note >> D_0, note
            elif i == 3:
                assert note >> E_0, note
            elif i == 4:
                assert note >> F_0, note
            elif i == 5:
                assert note >> G_0, note
            elif i == 6:
                assert note >> A_0, note
            elif i == 7:
                assert note >> B_0, note
            elif i == 8:
                assert note >> C_1, note
            elif i == 9:
                assert note >> D_1, note
            elif i == 10:
                assert note >> E_1, note
            elif i == 11:
                assert note >> F_1, note
            elif i == 12:
                assert note >> G_1, note
            elif i == 13:
                assert note >> A_1, note
            elif i == 14:
                assert note >> B_1, note
            elif i == 15:
                assert note >> C_2, note
            # ..............................
            elif i == 64:
                assert note >> C_9, note
            elif i == 65:
                assert note >> D_9, note
            elif i == 66:
                assert note >> E_9, note
            elif i == 67:
                assert note >> F_9, note
            elif i == 68:
                assert note >> G_9, note
            elif i == 69:
                assert note >> A_9, note
            elif i == 70:
                assert note >> B_9, note
            elif i == 71:
                assert note >> NotePitch('C', '', 10), note


    def testBMajorScaleGenerator(self):

        octaves_to_test = 18
        intervals = 7
        notes_to_test = octaves_to_test * intervals + 1  # 18 * 7 + 1 = 127

        for i, note in enumerate(
            self.b_major.spell(
                note_count=notes_to_test, yield_all=False
            )
        ):
            i += 1
            if i == 1:
                assert note >> B_0, note
            elif i == 2:
                assert note >> C_SHARP_1, note
            elif i == 3:
                assert note >> D_SHARP_1, note
            elif i == 4:
                assert note >> E_1, note
            elif i == 5:
                assert note >> F_SHARP_1, note
            elif i == 6:
                assert note >> G_SHARP_1, note
            elif i == 7:
                assert note >> A_SHARP_1, note
            elif i == 8:
                assert note >> B_1, note
            elif i == 9:
                assert note >> C_SHARP_2, note
            elif i == 10:
                assert note >> D_SHARP_2, note
            elif i == 11:
                assert note >> E_2, note
            elif i == 12:
                assert note >> F_SHARP_2, note
            elif i == 13:
                assert note >> G_SHARP_2, note
            elif i == 14:
                assert note >> A_SHARP_2, note
            elif i == 15:
                assert note >> B_2, note
            elif i == 16:
                assert note >> C_SHARP_3, note
            # ..............................
            elif i == 64:
                assert note >> B_9, note
            elif i == 65:
                assert note >> NotePitch('C', '#', 10), note


    def testDFlatMajorScaleGenerator(self):

        octaves_to_test = 18
        intervals = 7
        notes_to_test = octaves_to_test * intervals + 1  # 18 * 7 + 1 = 127

        for i, note in enumerate(
            self.d_flat_major.spell(
                note_count=notes_to_test, yield_all=False
            )
        ):
            i += 1
            if i == 1:
                assert note >> D_FLAT_0, note
            elif i == 2:
                assert note >> E_FLAT_0, note
            elif i == 3:
                assert note >> F_0, note
            elif i == 4:
                assert note >> G_FLAT_0, note
            elif i == 5:
                assert note >> A_FLAT_0, note
            elif i == 6:
                assert note >> B_FLAT_0, note
            elif i == 7:
                assert note >> C_1, note
            elif i == 8:
                assert note >> D_FLAT_1, note
            elif i == 9:
                assert note >> E_FLAT_1, note
            elif i == 10:
                assert note >> F_1, note
            elif i == 11:
                assert note >> G_FLAT_1, note
            elif i == 12:
                assert note >> A_FLAT_1, note
            elif i == 13:
                assert note >> B_FLAT_1, note
            elif i == 14:
                assert note >> C_2, note
            elif i == 15:
                assert note >> D_FLAT_2, note
            elif i == 16:
                assert note >> E_FLAT_2, note
            # ..............................
            elif i == 64:
                assert note >> D_FLAT_9, note
            elif i == 65:
                assert note >> E_FLAT_9, note
            elif i == 66:
                assert note >> F_9, note
            elif i == 67:
                assert note >> G_FLAT_9, note
            elif i == 68:
                assert note >> A_FLAT_9, note
            elif i == 69:
                assert note >> B_FLAT_9, note
            elif i == 70:
                assert note >> NotePitch('C', '', 10), note


class TonalScaleSpellMethodTest(unittest.TestCase):

    def setUp(self):
        print()
        self.scales = {
            'Ab_chromatic': ChromaticScale('A', 'b'),
            'B_major': MajorScale('B'),
            'Bb_minor': MinorScale('B', 'b'),
            'C_mel_minor': MelodicMinorScale('C'),
            'F#_har_minor': HarmonicMinorScale('F', '#'),
        }

    def testNoteCount(self):
        ''' tests yielded note count is what has been required '''
        for key in self.scales.values():
            max_notes = random.randint(2, 64)
            # max_notes = 65
            print(
                'Testing {}{} {}._count_notes( note_count=1..{} ) ...'.format(
                    key.root.chr, key.root.repr_alt, key.name(), max_notes
                )
            )
            for count in range(max_notes):
                count += 1
                yielded_notes = len(
                    [ n for n in key._count_notes(
                        note_count=count, start_note=None, yield_all=False
                    ) ]
                )
                assert yielded_notes == count, (yielded_notes, count)


    def testDiatonicStartNote(self):
        ''' tests that first yielded note == diatonic start_note '''
        for key in self.scales.values():
            d = random.randint(2, 64)
            print(
                'Testing {}{} {}._count_notes( start_note = note({}) ) ...'.format(
                    key.root.chr, key.root.repr_alt, key.name(), d
                )
            )
            for note in key._count_notes(
                note_count=1, start_note=key[d], yield_all=True
            ):
                assert note >> NotePitch(*key[d]), note


    def testNonDiatonicStartNoteYieldNotes(self):
        ''' tests 1st yielded item == expected diatonic note when:
                * start_note is non-diatonic to the scale
                * Nones ARE NOT being yielded
        '''
        test_parameters = [

            {
                'scale': self.scales['Ab_chromatic'],
                'non_diatonic_note': A_SHARP_1, # enharmonic
                'exp_diatonic_note': B_FLAT_1, # equals
            },

            {
                'scale': self.scales['B_major'],
                'non_diatonic_note': D_1,      # missing note
                'exp_diatonic_note': D_SHARP_1, # next note
            },

            {
                'scale': self.scales['Bb_minor'],
                'non_diatonic_note': D_SHARP_1, # enharmonic
                'exp_diatonic_note': E_FLAT_1, # equals
            },
            {
                'scale': self.scales['C_mel_minor'],
                'non_diatonic_note': F_SHARP_0, # missing note
                'exp_diatonic_note': G_0, # next note
            },
            {
                'scale': self.scales['F#_har_minor'],
                'non_diatonic_note': C_FLAT_4, # enharmonic
                'exp_diatonic_note': B_3, # equals
            },

        ]

        for param in test_parameters:
            key = param['scale']
            non = param['non_diatonic_note']
            exp = param['exp_diatonic_note']
            print(
                'Testing {}{} {}._count_notes( start_note = non_diatonic_note , yield_all = 0 ) ...'.format(
                        key.root.chr, key.root.repr_alt, key.name()
                )
            )
            for note in key._count_notes(
                note_count=1, start_note=non, yield_all=False
            ):
                assert note != None, type(note)         # yield all=False ensures no Nones
                assert note >> NotePitch(*exp), (note, exp)


    def testNonDiatonicStartNoteYieldAll(self):
        ''' tests 1st yielded item == expected value when:
                * start_note is non-diatonic to the scale
                * Nones ARE being yielded
        '''
        test_parameters = [
            {
                'scale': self.scales['Ab_chromatic'],
                'non_diatonic_note': A_SHARP_1, # enharmonic
                'exp_diatonic_note': B_FLAT_1, # equals
            },
            {
                'scale': self.scales['B_major'],
                'non_diatonic_note': D_1,      # missing note
                'exp_diatonic_note': None,              # yields a None, not D#
            },
            {
                'scale': self.scales['Bb_minor'],
                'non_diatonic_note': D_SHARP_1, # enharmonic
                'exp_diatonic_note': E_FLAT_1, # equals
            },
            {
                'scale': self.scales['C_mel_minor'],
                'non_diatonic_note': F_SHARP_0, # missing note
                'exp_diatonic_note': None,              # yields a None, not G
            },
            {
                'scale': self.scales['F#_har_minor'],
                'non_diatonic_note': C_FLAT_4, # enharmonic
                'exp_diatonic_note': B_3, # equals
            },
        ]

        for param in test_parameters:
            key = param['scale']
            non = param['non_diatonic_note']
            exp = param['exp_diatonic_note']

            print(
                'Testing {}{} {}._count_notes( start_note = non_diatonic_note , yield_all = 1 ) ...'.format(
                    key.root.chr, key.root.repr_alt, key.name()
                )
            )

            for note in key._count_notes(
                note_count=1, start_note=non, yield_all=True
            ):
                if note == None:
                    assert note == exp, (note, exp)
                else:
                    assert note >> exp, (note, exp)

                break # test only first yielded value even if not a note


    def testMajorNoneYields(self):
        for i, note in enumerate(MajorScale('C')._count_notes(
            note_count=64, start_note=None, yield_all=True
        )):
            if i == 0:
                assert note >> C_0, note
            elif i == 1:
                assert note == None
            elif i == 2:
                assert note >> D_0, note
            elif i == 3:
                assert note == None
            elif i == 4:
                assert note >> E_0, note
            elif i == 5:
                assert note >> F_0, note
            elif i == 6:
                assert note == None
            elif i == 7:
                assert note >> G_0, note
            elif i == 8:
                assert note == None
            elif i == 9:
                assert note >> A_0, note
            elif i == 10:
                assert note == None
            elif i == 11:
                assert note >> B_0, note
            elif i == 12:
                assert note >> C_1, note


    def testChromaticNoneYields(self):
        ''' tests Chromatic scale does NEVER yield None even with yield_all '''
        for i, note in enumerate(self.scales['Ab_chromatic']._count_notes(
            note_count=128, start_note=None, yield_all=True
        )):
            assert note != None


    def testDegreeOrderOverOct(self):
        for i, note in enumerate(
            MajorScale('A').spell(
            note_count=6,
            start_note=None,
            yield_all=True,
        )):
            i += 1
            if i == 1:
                assert note >> A_0, note
            elif i == 2:
                assert not note, note
            elif i == 3:
                assert note >> B_0, note
            elif i == 4:
                assert not note, note
            elif i == 5:
                assert note >> C_SHARP_1, note
            elif i == 6:
                assert note >> D_1, note
            elif i == 7:
                assert not note, note
            elif i == 8:
                assert note >> E_1, note



    def testCustomDegrees(self):
        '''
        B is not part of C major pentatonic =>
            * 1st note should be None
            * 2nd note should be C
        '''
        for i, note in enumerate(
            MajorPentatonicScale('C').spell(
                yield_all=1,
                start_note=B_3,
                note_count=3
            )
        ):
            i += 1
            if i == 1:
                assert not note, note
            if i == 2:
                assert note, note
                assert note >> C_4, note


    def testMaxOctave(self):
        try:
            note = NotePitch('C', NotePitch.MAXIMUM_OCTAVE + 1)
        except InvalidOctave:
            note = None
        finally:
            assert not note




from ..pitch import NotePitch


C_SHARP_0 = NotePitch('C', '#' , 0)
D_SHARP_0 = NotePitch('D', '#' , 0)
E_SHARP_0 = NotePitch('E', '#' , 0)
F_SHARP_0 = NotePitch('F', '#' , 0)
G_SHARP_0 = NotePitch('G', '#' , 0)
A_SHARP_0 = NotePitch('A', '#' , 0)
B_SHARP_0 = NotePitch('B', '#' , 0)

C_SHARP_1 = NotePitch('C', '#' , 1)
D_SHARP_1 = NotePitch('D', '#' , 1)
E_SHARP_1 = NotePitch('E', '#' , 1)
F_SHARP_1 = NotePitch('F', '#' , 1)
G_SHARP_1 = NotePitch('G', '#' , 1)
A_SHARP_1 = NotePitch('A', '#' , 1)
B_SHARP_1 = NotePitch('B', '#' , 1)

C_SHARP_2 = NotePitch('C', '#' , 2)
D_SHARP_2 = NotePitch('D', '#' , 2)
E_SHARP_2 = NotePitch('E', '#' , 2)
F_SHARP_2 = NotePitch('F', '#' , 2)
G_SHARP_2 = NotePitch('G', '#' , 2)
A_SHARP_2 = NotePitch('A', '#' , 2)
B_SHARP_2 = NotePitch('B', '#' , 2)

C_SHARP_3 = NotePitch('C', '#' , 3)
D_SHARP_3 = NotePitch('D', '#' , 3)
E_SHARP_3 = NotePitch('E', '#' , 3)
F_SHARP_3 = NotePitch('F', '#' , 3)
G_SHARP_3 = NotePitch('G', '#' , 3)
A_SHARP_3 = NotePitch('A', '#' , 3)
B_SHARP_3 = NotePitch('B', '#' , 3)

C_SHARP_4 = NotePitch('C', '#' , 4)
D_SHARP_4 = NotePitch('D', '#' , 4)
E_SHARP_4 = NotePitch('E', '#' , 4)
F_SHARP_4 = NotePitch('F', '#' , 4)
G_SHARP_4 = NotePitch('G', '#' , 4)
A_SHARP_4 = NotePitch('A', '#' , 4)
B_SHARP_4 = NotePitch('B', '#' , 4)

C_SHARP_5 = NotePitch('C', '#' , 5)
D_SHARP_5 = NotePitch('D', '#' , 5)
E_SHARP_5 = NotePitch('E', '#' , 5)
F_SHARP_5 = NotePitch('F', '#' , 5)
G_SHARP_5 = NotePitch('G', '#' , 5)
A_SHARP_5 = NotePitch('A', '#' , 5)
B_SHARP_5 = NotePitch('B', '#' , 5)

C_SHARP_6 = NotePitch('C', '#' , 6)
D_SHARP_6 = NotePitch('D', '#' , 6)
E_SHARP_6 = NotePitch('E', '#' , 6)
F_SHARP_6 = NotePitch('F', '#' , 6)
G_SHARP_6 = NotePitch('G', '#' , 6)
A_SHARP_6 = NotePitch('A', '#' , 6)
B_SHARP_6 = NotePitch('B', '#' , 6)

C_SHARP_7 = NotePitch('C', '#' , 7)
D_SHARP_7 = NotePitch('D', '#' , 7)
E_SHARP_7 = NotePitch('E', '#' , 7)
F_SHARP_7 = NotePitch('F', '#' , 7)
G_SHARP_7 = NotePitch('G', '#' , 7)
A_SHARP_7 = NotePitch('A', '#' , 7)
B_SHARP_7 = NotePitch('B', '#' , 7)

C_SHARP_8 = NotePitch('C', '#' , 8)
D_SHARP_8 = NotePitch('D', '#' , 8)
E_SHARP_8 = NotePitch('E', '#' , 8)
F_SHARP_8 = NotePitch('F', '#' , 8)
G_SHARP_8 = NotePitch('G', '#' , 8)
A_SHARP_8 = NotePitch('A', '#' , 8)
B_SHARP_8 = NotePitch('B', '#' , 8)

C_SHARP_9 = NotePitch('C', '#' , 9)
D_SHARP_9 = NotePitch('D', '#' , 9)
E_SHARP_9 = NotePitch('E', '#' , 9)
F_SHARP_9 = NotePitch('F', '#' , 9)
G_SHARP_9 = NotePitch('G', '#' , 9)
A_SHARP_9 = NotePitch('A', '#' , 9)
B_SHARP_9 = NotePitch('B', '#' , 9)



import unittest

from . import NotePitch

from .constants import (
    F_4, B_4, B_5,
    C_FLAT_3, C_FLAT_4, D_FLAT_4, B_FLAT_4, E_FLAT_5, D_FLAT_3, 
    E_SHARP_3, B_SHARP_3, D_SHARP_4, C_SHARP_5, C_SHARP_3,
    D_DOUBLE_FLAT_4,
)

from ..errors import InvalidNote

__all__ = [
    'TestInvalidNotes',
    'NoteEqualityTest',
]


class TestInvalidNotes(unittest.TestCase):

    def testInvalidChars(self):
        self.assertRaises(
            InvalidNote, lambda : NotePitch('Y')
        )


class NoteEqualityTest(unittest.TestCase):

    DANGEROUS_NON_EQUALS = (
        # ''' Used mainly to test B#, Cb, etc... '''

        (C_FLAT_3, B_SHARP_3),
        (E_FLAT_5, D_SHARP_4),

        (B_5, C_FLAT_4),

        (E_SHARP_3, F_4),


        (NotePitch('B', 'b'), NotePitch('C', 'bb')),
        (NotePitch('A', '#'), NotePitch('C', 'bb')),

        (NotePitch('B'     ), NotePitch('C', 'b')),
        (NotePitch('A', '##'), NotePitch('C', 'b')),

        (NotePitch('B', '#'), NotePitch('C', '')),
        (NotePitch('B', '#'), NotePitch('D', 'bb')),

        (NotePitch('B', '##'), NotePitch('C', '#')),
        (NotePitch('B', '##'), NotePitch('D', 'b')),


        # these should eval False OK
        (NotePitch('A', '#'), B_FLAT_4),
        (NotePitch('A', '##'), B_4),
        (NotePitch('C', ''), D_DOUBLE_FLAT_4),
        (NotePitch('C', '#'), D_FLAT_4),
    )

    def setUp(self):
        print()

    def testNotEnharmonic(self):
        for note_pair in self.DANGEROUS_NON_EQUALS:
            assert not note_pair[0] >> note_pair[1], (note_pair[0], note_pair[1])
            assert note_pair[0] - note_pair[1] != 0, (note_pair[0], note_pair[1])


    def testAreEnharmonic(self):
        ''' checks note_pairs in enharmonic rows for different types of enharmonic equality '''
        for note_pair in NotePitch.EnharmonicMatrix():

            # PASS ENHARMONIC EQUALITY
            assert note_pair[0] == note_pair[1], (note_pair[0], note_pair[1])

            # DO NOT PASS NOTE EQUALITY
            assert not note_pair[0] >> note_pair[1], (note_pair[0], note_pair[1])
            assert not note_pair[0] ** note_pair[1], (note_pair[0], note_pair[1])

            # CHECK DELTA_ST == 0
            assert not note_pair[0] < note_pair[1], (note_pair[0], note_pair[1])
            assert not note_pair[0] > note_pair[1], (note_pair[0], note_pair[1])
            assert note_pair[0] <= note_pair[1], (note_pair[0], note_pair[1])
            assert note_pair[0] >= note_pair[1], (note_pair[0], note_pair[1])
            assert note_pair[0] - note_pair[1] == 0, (note_pair[0], note_pair[1])


    def testEnharmonicOperators(self):
        ''' '''
        print('Testing enharmonic operators')
        assert C_SHARP_3 ** C_SHARP_5     # loosest equality
        assert C_SHARP_3 == D_FLAT_3      # enhamonic notes
        assert not D_FLAT_3 >> C_SHARP_3  # enh note
        assert not C_SHARP_3 >> D_FLAT_3  # enh note

        # ** same note, ignr oct
        # == enhr note, same oct
        # >> same note, same oct

        # // enhr note, ignr oct   do I need to implement this operator ?



#!/usr/bin/env  python3

"""
this script displays a command-line representation of your instrument's fretboard, tuned to your liking
note patterns will be displayed for any given mode (scale/chord) for any given root note
the tunings directory already contains some pre-defined instrument tunings in the form of .json files
modify them or add your own and they will automaticaly become available at runtime
"""

import sys
import argparse

from kord.keys.scales import (
    MajorScale, MinorScale, MelodicMinorScale, HarmonicMinorScale,
    MajorPentatonicScale, MinorPentatonicScale,
    AugmentedScale, DiminishedScale,
    IonianMode, LydianMode, MixolydianMode,
    AeolianMode, DorianMode, PhrygianMode, LocrianMode,
    ChromaticScale,
)

from kord.keys.chords import (
    PowerChord,
    MajorTriad, MinorTriad, AugmentedTriad, DiminishedTriad,
    MajorSeventhChord, MinorSeventhChord, DominantSeventhChord,
    DiminishedSeventhChord, HalfDiminishedSeventhChord,
    DominantNinthChord, DominantMinorNinthChord,
    MajorNinthChord, MinorNinthChord,
    MajorSixthChord, MinorSixthChord,
    Suspended4Chord, Suspended2Chord,
)


from kord.notes import NotePitch

from kord.instruments import PluckedStringInstrument, max_frets_on_screen

from kord.errors import InvalidInstrument, InvalidNote, InvalidAlteration

from bestia.output import echo

import tuner

AVAILABLE_SCALES = {
    scale.notations[0]: scale for scale in (
        MajorScale,
        MinorScale,
        MelodicMinorScale,
        HarmonicMinorScale,
        MajorPentatonicScale,
        MinorPentatonicScale,
        AugmentedScale,
        DiminishedScale,
        IonianMode,
        LydianMode,
        MixolydianMode,
        AeolianMode,
        DorianMode,
        PhrygianMode,
        LocrianMode,
        ChromaticScale,
    )
}

AVAILABLE_CHORDS = {
    chord.notations[0]: chord for chord in (
        PowerChord,
        Suspended4Chord,
        Suspended2Chord,
        MajorTriad,
        MinorTriad,
        AugmentedTriad,
        DiminishedTriad,
        MajorSixthChord,
        MinorSixthChord,
        MajorSeventhChord,
        MinorSeventhChord,
        DominantSeventhChord,
        DiminishedSeventhChord,
        HalfDiminishedSeventhChord,
        DominantNinthChord,
        DominantMinorNinthChord,
        MajorNinthChord,
        MinorNinthChord,
    )
}

TUNINGS = tuner.load_tuning_data()

def parse_arguments():
    parser = argparse.ArgumentParser(
        description='<<< Fretboard visualizer tool for the kord framework >>>',
    )
    parser.add_argument(
        'root',
        help='select key ROOT note',
    )

    parser.add_argument(
        '-d', '--degrees',
        help='show degree numbers instead of note pitches',
        action='store_true',
    )

    mode_group = parser.add_mutually_exclusive_group()

    scale_choices = list( AVAILABLE_SCALES.keys() )
    scale_choices.sort()
    mode_group.add_argument(
        '-s', '--scale',
        help='{}'.format(str(scale_choices).lstrip('[').rstrip(']').replace('\'', '')),
        choices=scale_choices,
        metavar='',
    )

    chord_choices = list( AVAILABLE_CHORDS.keys() )
    chord_choices.sort()
    mode_group.add_argument(
        '-c', '--chord',
        help='{}'.format(str(chord_choices).lstrip('[').rstrip(']').replace('\'', '')),
        choices=chord_choices,
        default=MajorTriad.notations[0],
        metavar='',
    )

    instr_choices = list( TUNINGS.keys() )
    parser.add_argument(
        '-i', '--instrument',
        help='{}'.format(str(instr_choices).lstrip('[').rstrip(']').replace('\'', '')),
        choices=instr_choices,
        default='guitar',
        metavar='',
    )

    parser.add_argument(
        '-t', '--tuning',
        help='check .json files for available options',
        default='standard',
        metavar='',
    )

    parser.add_argument(
        '-f', '--frets',
        help='1, 2, .., {}'.format(PluckedStringInstrument.maximum_frets()),
        choices=[ f+1 for f in range(PluckedStringInstrument.maximum_frets()) ],
        default=max_frets_on_screen(),
        metavar='',
        type=int,
    )

    parser.add_argument(
        '-v', '--verbosity',
        help='0, 1, 2',
        choices= (0, 1, 2),
        default=1,
        metavar='',
        type=int,
    )

    args = parser.parse_args()

    # some args require extra validation than what argparse can offer, let's do that here...
    try:
        # validate tuning
        if args.tuning not in TUNINGS[args.instrument].keys():
            raise InvalidInstrument(
                "fretboard.py: error: argument -t/--tuning: invalid choice: '{}' (choose from {}) ".format(
                    args.tuning,
                    str( list( TUNINGS[args.instrument].keys() ) ).lstrip('[').rstrip(']'),
                )
            )

        # validate root note
        note_chr = args.root[:1].upper()
        if note_chr not in NotePitch.possible_chars():
            raise InvalidNote(
                "fretboard.py: error: argument ROOT: invalid note: '{}' (choose from {}) ".format(
                    note_chr,
                    str( NotePitch.possible_chars() ).lstrip('[').rstrip(']')
                )
            )

        # validate root alteration
        note_alt = args.root[1:]
        if note_alt and note_alt not in NotePitch.input_alterations():
            raise InvalidAlteration(
                "fretboard.py: error: argument ROOT: invalid alteration: '{}' (choose from {}) ".format(
                    note_alt,
                    str( NotePitch.input_alterations() ).lstrip('(').rstrip(')')
                )
            )

        args.root = (note_chr, note_alt)

    except Exception as x:
        parser.print_usage()
        print(x)
        args = None

    return args

def print_mode(mode):
    clr = 'blue'
    mode = '{} {}'.format(
        str(mode.root)[:-1],
        mode.name(),
    )
    echo(mode, clr)

def print_instrument(instrument, tuning):
    clr = 'yellow'
    echo(instrument.title(), clr, 'underline', mode='raw')
    echo(': ', clr, mode='raw')
    echo('{} tuning'.format(tuning), clr)


def run(args):
    # default mode is chord, so use scale if set
    if args.scale:
        KeyMode = AVAILABLE_SCALES[args.scale]
    elif args.chord:
        KeyMode = AVAILABLE_CHORDS[args.chord]

    root = NotePitch(args.root[0], args.root[1])
    key_mode = KeyMode(*root)
    if not key_mode.validate():
        print(
            '{} {} cannot be visualized'.format(
                str(root)[:-1],
                key_mode.name(),
            )
        )
        return -1

    if args.verbosity:
        print_instrument(args.instrument, args.tuning)

    instrument = PluckedStringInstrument(
        * TUNINGS[args.instrument][args.tuning]
    )

    instrument.render_fretboard(
        mode=key_mode,
        frets=args.frets,
        verbose=args.verbosity,
        show_degrees=args.degrees,
    )

    if args.verbosity:
        print_mode(key_mode)

    return 0


if __name__ == '__main__':
    rc = -1
    valid_args = parse_arguments()
    if not valid_args:
        rc = 2
    else:
        rc = run(valid_args)
    sys.exit(rc)




from enum import IntEnum

__all__ = [
    'Intervals',
]

class Intervals(IntEnum):

    UNISON = 0
    DIMINISHED_SECOND = 0

    MINOR_SECOND = 1
    AUGMENTED_UNISON = 1

    MAJOR_SECOND = 2
    DIMINISHED_THIRD = 2

    MINOR_THIRD = 3
    AUGMENTED_SECOND = 3

    MAJOR_THIRD = 4
    DIMINISHED_FOURTH = 4

    PERFECT_FOURTH = 5
    AUGMENTED_THIRD = 5

    DIMINISHED_FIFTH = 6
    AUGMENTED_FOURTH = 6

    PERFECT_FIFTH = 7
    DIMINISHED_SIXTH = 7

    MINOR_SIXTH = 8
    AUGMENTED_FIFTH = 8

    MAJOR_SIXTH = 9
    DIMINISHED_SEVENTH = 9

    MINOR_SEVENTH = 10
    AUGMENTED_SIXTH = 10

    MAJOR_SEVENTH = 11
    DIMINISHED_OCTAVE = 11

    PERFECT_OCTAVE = 12
    AUGMENTED_SEVENTH = 12






from .pitch_parser import NotePitchParser

from ..notes import NotePitch
from ..keys.chords import *

from ..errors import InvalidNote, InvalidAlteration, InvalidOctave, InvalidChord

from bestia.output import echo

__all__ = [
    'ChordParser',
]


class ChordParser:
    ''' chord symbols still left to add:

        Fadd4

        Am11

        G13
        B7b9
    '''

    RECOGNIZED_CHORDS = (
        PowerChord, Suspended4Chord, Suspended2Chord,
        MajorTriad, MinorTriad, AugmentedTriad, DiminishedTriad,
        MajorSixthChord, MinorSixthChord,
        MajorSeventhChord, MinorSeventhChord, DominantSeventhChord,
        HalfDiminishedSeventhChord, DiminishedSeventhChord,
        MajorAdd9Chord, MinorAdd9Chord, AugmentedAdd9Chord, DiminishedAdd9Chord,
        DominantNinthChord, DominantMinorNinthChord,
        MajorNinthChord, MinorNinthChord,
    )

    BASS_NOTE_SEP = '/'

    def __init__(self, symbol):
        self.symbol = symbol.strip()
        self.reset()


    def reset(self):
        self.root = None
        self.flavor = None
        self.to_parse = self.symbol.replace(' ', '')  # 'B aug7'


    @property
    def is_inverted(self) -> bool:
        return self.BASS_NOTE_SEP in self.symbol

    @property
    def bass(self):
        if not self.is_inverted:
            return self.root
        return NotePitchParser(
            self.symbol.split(self.BASS_NOTE_SEP)[-1]
        ).parse()


    def _parse_root(self):
        ''' this should NOT validate anything!
            only guess how many of the symbol's first 3 chars make up the root
            return value will be validated using NotePitchParser::parse()
        '''
        possible_root = self.to_parse[:3]
        if len(possible_root) in (0, 1,):
            # symbol='', possible_root=''
            # symbol='D', possible_root='D'
            return possible_root

        if possible_root[1:] in ('bb', '##', '♯♯', '♭♭'):
            # symbol='Dbbmin7', possible_root='Dbb'
            return possible_root[:3]

        if possible_root[1] in ('b', '#', '♯', '♭', '𝄫', '𝄪'):
            # symbol='Dbmin7', possible_root='Db'
            return possible_root[:2]

        # symbol='Dmin7', possible_root='D'
        return possible_root[:1]


    def _parse_flavor(self):
        # shouldnt  this go in a _parse_alt() function
        if 'sharp' in self.to_parse.lower() or 'flat' in self.to_parse.lower():
            raise InvalidAlteration(self.symbol)

        for chord_class in self.RECOGNIZED_CHORDS:
            if self.to_parse in chord_class.notations:
                return chord_class
        return None


    def parse(self):
        ''' - sure   chord is correct?  return TonalKey()
            - unsure chord is correct?  return None
            - sure   chord is wrong  ?  raise  InvalidChord()
        '''
        try:
            # parse root out of first 3 chars
            root = self._parse_root()
            self.root = NotePitchParser(root).parse()
            # print(self.root)

            # ignore Chord/Bass notes on inverted chords
            if self.is_inverted:
                self.to_parse = self.to_parse.split(self.BASS_NOTE_SEP)[0]

            # parse flavor out of remaining to parse
            self.to_parse = self.to_parse[len(root):]
            self.flavor = self._parse_flavor()

        except Exception:
            # echo(self.symbol, 'red')
            raise InvalidChord(self.symbol)

        # c = 'cyan'
        # if not self.symbol or not self.flavor:
        #     c = 'red'
        # echo(
        #     f'{self.symbol} = {self.root} {self.flavor} {self.bass}',
        #     c,
        # )
        if self.flavor and self.root:
            # init instance of Chord class using Chord root
            return self.flavor(*self.root)
        return None




import unittest

from .chords import (
    PowerChord,
    Suspended4Chord, Suspended2Chord,
    MajorTriad, MinorTriad, AugmentedTriad, DiminishedTriad,
    MajorSixthChord, MinorSixthChord,
    MajorSeventhChord, MinorSeventhChord, DominantSeventhChord,
    DiminishedSeventhChord, HalfDiminishedSeventhChord,
    MajorAdd9Chord, MinorAdd9Chord, AugmentedAdd9Chord, DiminishedAdd9Chord,
    MajorNinthChord, MinorNinthChord,
    DominantNinthChord, DominantMinorNinthChord,
)

from ..notes.constants import (
    A_0, A_1, A_2, B_0, B_1, C_0, C_1, C_2, C_3, D_0, D_1, E_0, E_1, F_0, F_1, G_0, G_1,
    E_FLAT_0, E_FLAT_1, B_FLAT_0, B_FLAT_1, G_FLAT_0, G_FLAT_1, D_FLAT_0, D_FLAT_1,
    G_SHARP_0, G_SHARP_1,
    B_DOUBLE_FLAT_0, B_DOUBLE_FLAT_1,
)

from ..errors import InvalidNote, InvalidAlteration, InvalidOctave, InvalidChord

__all__ = [
    'ChordTest',
]


class ChordTest(unittest.TestCase):

    def testPowerChord(self):
        chord = PowerChord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert not chord[3] , chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert not chord[10] , chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testSuspended4Chord(self):
        chord = Suspended4Chord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert not chord[3] , chord[3]
        assert chord[4] >> F_0, chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert not chord[10] , chord[10]
        assert chord[11] >> F_1, chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testSuspended2Chord(self):
        chord = Suspended2Chord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert chord[2] >> D_0, chord[2]
        assert not chord[3] , chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert chord[9] >> D_1, chord[9]
        assert not chord[10] , chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testMajorTriad(self):
        chord = MajorTriad(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert chord[3] >> E_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert chord[10] >> E_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testMinorTriad(self):
        chord = MinorTriad(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert chord[3] >> E_FLAT_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert chord[10] >> E_FLAT_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testAugmentedTriad(self):
        chord = AugmentedTriad(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert chord[3] >> E_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_SHARP_0, chord[5]
        assert not chord[6] , chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert chord[10] >> E_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_SHARP_1, chord[12]
        assert not chord[13] , chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testDiminishedTriad(self):
        chord = DiminishedTriad(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert chord[3] >> E_FLAT_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_FLAT_0, chord[5]
        assert not chord[6] , chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert chord[10] >> E_FLAT_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_FLAT_1, chord[12]
        assert not chord[13] , chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testMajorSixthChordChord(self):
        chord = MajorSixthChord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert chord[3] >> E_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert chord[6] >> A_0, chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert chord[10] >> E_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert chord[13] >> A_1, chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testMinorSixthChord(self):
        chord = MinorSixthChord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert chord[3] >> E_FLAT_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert chord[6] >> A_0, chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert chord[10] >> E_FLAT_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert chord[13] >> A_1, chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testMajorSeventhChord(self):
        chord = MajorSeventhChord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert chord[3] >> E_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert chord[7] >> B_0, chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert chord[10] >> E_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert chord[14] >> B_1, chord[14]
        assert chord[15] >> C_2, chord[15]


    def testMinorSeventhChord(self):
        chord = MinorSeventhChord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert chord[3] >> E_FLAT_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert chord[7] >> B_FLAT_0, chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert chord[10] >> E_FLAT_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert chord[14] >> B_FLAT_1, chord[14]
        assert chord[15] >> C_2, chord[15]


    def testDominantSeventhChord(self):
        chord = DominantSeventhChord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert chord[3] >> E_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert chord[7] >> B_FLAT_0, chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert chord[10] >> E_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert chord[14] >> B_FLAT_1, chord[14]
        assert chord[15] >> C_2, chord[15]


    def testHalfDiminishedSeventhChord(self):
        chord = HalfDiminishedSeventhChord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert chord[3] >> E_FLAT_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_FLAT_0, chord[5]
        assert not chord[6] , chord[6]
        assert chord[7] >> B_FLAT_0, chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert chord[10] >> E_FLAT_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_FLAT_1, chord[12]
        assert not chord[13] , chord[13]
        assert chord[14] >> B_FLAT_1, chord[14]
        assert chord[15] >> C_2, chord[15]


    def testDiminishedSeventhChord(self):
        chord = DiminishedSeventhChord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert not chord[2] , chord[2]
        assert chord[3] >> E_FLAT_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_FLAT_0, chord[5]
        assert not chord[6] , chord[6]
        assert chord[7] >> B_DOUBLE_FLAT_0, chord[7]
        assert chord[8] >> C_1, chord[8]
        assert not chord[9] , chord[9]
        assert chord[10] >> E_FLAT_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_FLAT_1, chord[12]
        assert not chord[13] , chord[13]
        assert chord[14] >> B_DOUBLE_FLAT_1, chord[14]
        assert chord[15] >> C_2, chord[15]


    def testMajorAdd9Chord(self):
        chord = MajorAdd9Chord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert chord[2] >> D_0, chord[2]
        assert chord[3] >> E_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert chord[9] >> D_1, chord[9]
        assert chord[10] >> E_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testMinorAdd9Chord(self):
        chord = MinorAdd9Chord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert chord[2] >> D_0, chord[2]
        assert chord[3] >> E_FLAT_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert chord[9] >> D_1, chord[9]
        assert chord[10] >> E_FLAT_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testAugmentedAdd9Chord(self):
        chord = AugmentedAdd9Chord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert chord[2] >> D_0, chord[2]
        assert chord[3] >> E_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_SHARP_0, chord[5]
        assert not chord[6] , chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert chord[9] >> D_1, chord[9]
        assert chord[10] >> E_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_SHARP_1, chord[12]
        assert not chord[13] , chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testDiminishedAdd9Chord(self):
        chord = DiminishedAdd9Chord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert chord[2] >> D_0, chord[2]
        assert chord[3] >> E_FLAT_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_FLAT_0, chord[5]
        assert not chord[6] , chord[6]
        assert not chord[7] , chord[7]
        assert chord[8] >> C_1, chord[8]
        assert chord[9] >> D_1, chord[9]
        assert chord[10] >> E_FLAT_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_FLAT_1, chord[12]
        assert not chord[13] , chord[13]
        assert not chord[14] , chord[14]
        assert chord[15] >> C_2, chord[15]


    def testMajorNinthChord(self):
        chord = MajorNinthChord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert chord[2] >> D_0, chord[2]
        assert chord[3] >> E_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert chord[7] >> B_0, chord[7]
        assert chord[8] >> C_1, chord[8]
        assert chord[9] >> D_1, chord[9]
        assert chord[10] >> E_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert chord[14] >> B_1, chord[14]
        assert chord[15] >> C_2, chord[15]


    def testMinorNinthChord(self):
        chord = MinorNinthChord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert chord[2] >> D_0, chord[2]
        assert chord[3] >> E_FLAT_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert chord[7] >> B_FLAT_0, chord[7]
        assert chord[8] >> C_1, chord[8]
        assert chord[9] >> D_1, chord[9]
        assert chord[10] >> E_FLAT_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert chord[14] >> B_FLAT_1, chord[14]
        assert chord[15] >> C_2, chord[15]


    def testDominantNinthChord(self):
        chord = DominantNinthChord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert chord[2] >> D_0, chord[2]
        assert chord[3] >> E_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert chord[7] >> B_FLAT_0, chord[7]
        assert chord[8] >> C_1, chord[8]
        assert chord[9] >> D_1, chord[9]
        assert chord[10] >> E_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert chord[14] >> B_FLAT_1, chord[14]
        assert chord[15] >> C_2, chord[15]


    def testDominantMinorNinthChord(self):
        chord = DominantMinorNinthChord(*C_3)
        assert not chord[0] , chord[0]
        assert chord[1] >> C_0, chord[1]
        assert chord[2] >> D_FLAT_0, chord[2]
        assert chord[3] >> E_0, chord[3]
        assert not chord[4] , chord[4]
        assert chord[5] >> G_0, chord[5]
        assert not chord[6] , chord[6]
        assert chord[7] >> B_FLAT_0, chord[7]
        assert chord[8] >> C_1, chord[8]
        assert chord[9] >> D_FLAT_1, chord[9]
        assert chord[10] >> E_1, chord[10]
        assert not chord[11] , chord[11]
        assert chord[12] >> G_1, chord[12]
        assert not chord[13] , chord[13]
        assert chord[14] >> B_FLAT_1, chord[14]
        assert chord[15] >> C_2, chord[15]
